from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Helper.collections_ import Option_fromValueWithDefault
from ...arctrl_core.Process.factor_value import FactorValue
from ...arctrl_core.Process.material_attribute_value import MaterialAttributeValue
from ...arctrl_core.Process.sample import Sample
from ...arctrl_core.Process.source import Source
from ...fable_library.list import (FSharpList, map, empty, append, choose, singleton, of_array, is_empty, head, tail)
from ...fable_library.option import (default_arg, map as map_1)
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import map as map_2
from ...fable_library.string_ import replace
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (string, object, IOptionalGetter, IRequiredGetter, unit, list_1 as list_1_3, IGetters)
from ...thoth_json_core.encode import list_1 as list_1_2
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1, ErrorReason_1, IDecoderHelpers_1)
from ..context.rocrate.isa_sample_context import context_jsonvalue
from ..decode import (Decode_uri, Decode_objectNoAdditionalProperties)
from ..encode import (try_include, try_include_list, try_include_list_opt)
from ..idtable import encode
from .factor_value import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_2)
from .material_attribute_value import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1)
from .source import (ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_3)

__A_ = TypeVar("__A_")

def ROCrate_genID(s: Sample) -> str:
    match_value: str | None = s.ID
    if match_value is None:
        match_value_1: str | None = s.Name
        if match_value_1 is None:
            return "#EmptySample"

        else: 
            return "#Sample_" + replace(match_value_1, " ", "_")


    else: 
        return match_value



def ROCrate_encoder(oa: Sample) -> IEncodable:
    additional_properties: FSharpList[IEncodable]
    list_4: FSharpList[IEncodable] = map(ROCrate_encoder_1, default_arg(oa.Characteristics, empty()))
    additional_properties = append(map(ROCrate_encoder_2, default_arg(oa.FactorValues, empty())), list_4)
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map_1(mapping, tupled_arg[1])

    def _arrow2623(__unit: None=None, oa: Any=oa) -> IEncodable:
        value_2: str = ROCrate_genID(oa)
        class ObjectExpr2622(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value_2)

        return ObjectExpr2622()

    class ObjectExpr2624(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Sample")

    class ObjectExpr2625(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_2.encode_string("Sample")

    def _arrow2627(value_5: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2626(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr2626()

    def _arrow2628(x: IEncodable, oa: Any=oa) -> IEncodable:
        return x

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2623()), ("@type", list_1_2(singleton(ObjectExpr2624()))), ("additionalType", ObjectExpr2625()), try_include("name", _arrow2627, oa.Name), try_include_list("additionalProperties", _arrow2628, additional_properties), ("@context", context_jsonvalue)]))
    class ObjectExpr2629(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2629()


class ObjectExpr2638(Decoder_1[tuple[MaterialAttributeValue | None, FactorValue | None]]):
    def Decode(self, s: IDecoderHelpers_1[Any], json: Any) -> FSharpResult_2[tuple[MaterialAttributeValue | None, FactorValue | None], tuple[str, ErrorReason_1[__A_]]]:
        def _arrow2635(__unit: None=None) -> str:
            match_value: FSharpResult_2[str, tuple[str, ErrorReason_1[__A_]]] = string.Decode(s, s.get_property("additionalType", json))
            return match_value.fields[0] if (match_value.tag == 0) else ""

        if (_arrow2635() if s.has_property("additionalType", json) else "") == "FactorValue":
            match_value_1: FSharpResult_2[FactorValue, tuple[str, ErrorReason_1[__A_]]] = ROCrate_decoder_1.Decode(s, json)
            return FSharpResult_2(1, match_value_1.fields[0]) if (match_value_1.tag == 1) else FSharpResult_2(0, (None, match_value_1.fields[0]))

        else: 
            match_value_2: FSharpResult_2[MaterialAttributeValue, tuple[str, ErrorReason_1[__A_]]] = ROCrate_decoder_2.Decode(s, json)
            return FSharpResult_2(1, match_value_2.fields[0]) if (match_value_2.tag == 1) else FSharpResult_2(0, (match_value_2.fields[0], None))



ROCrate_additionalPropertyDecoder: Decoder_1[tuple[MaterialAttributeValue | None, FactorValue | None]] = ObjectExpr2638()

def _arrow2645(get: IGetters) -> Sample:
    match_value: str | None
    object_arg: IOptionalGetter = get.Optional
    match_value = object_arg.Field("additionalType", Decode_uri)
    (pattern_matching_result,) = (None,)
    if match_value is None:
        pattern_matching_result = 0

    elif match_value == "Sample":
        pattern_matching_result = 0

    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 1:
        object_arg_1: IRequiredGetter = get.Required
        object_arg_1.Field("FailBecauseNotSample", unit)

    match_value_1: FSharpList[str] | None
    arg_5: Decoder_1[FSharpList[str]] = list_1_3(string)
    object_arg_2: IOptionalGetter = get.Optional
    match_value_1 = object_arg_2.Field("@type", arg_5)
    (pattern_matching_result_1,) = (None,)
    if match_value_1 is None:
        pattern_matching_result_1 = 0

    elif not is_empty(match_value_1):
        if head(match_value_1) == "Sample":
            if is_empty(tail(match_value_1)):
                pattern_matching_result_1 = 0

            else: 
                pattern_matching_result_1 = 1


        else: 
            pattern_matching_result_1 = 1


    else: 
        pattern_matching_result_1 = 1

    if pattern_matching_result_1 == 1:
        object_arg_3: IRequiredGetter = get.Required
        object_arg_3.Field("FailBecauseNotSample", unit)

    additional_properties: FSharpList[tuple[MaterialAttributeValue | None, FactorValue | None]] | None
    arg_9: Decoder_1[FSharpList[tuple[MaterialAttributeValue | None, FactorValue | None]]] = list_1_3(ROCrate_additionalPropertyDecoder)
    object_arg_4: IOptionalGetter = get.Optional
    additional_properties = object_arg_4.Field("additionalProperties", arg_9)
    pattern_input: tuple[FSharpList[MaterialAttributeValue] | None, FSharpList[FactorValue] | None]
    if additional_properties is not None:
        additional_properties_1: FSharpList[tuple[MaterialAttributeValue | None, FactorValue | None]] = additional_properties
        def chooser(tuple: tuple[MaterialAttributeValue | None, FactorValue | None]) -> MaterialAttributeValue | None:
            return tuple[0]

        def chooser_1(tuple_1: tuple[MaterialAttributeValue | None, FactorValue | None]) -> FactorValue | None:
            return tuple_1[1]

        pattern_input = (Option_fromValueWithDefault(empty(), choose(chooser, additional_properties_1)), Option_fromValueWithDefault(empty(), choose(chooser_1, additional_properties_1)))

    else: 
        pattern_input = (None, None)

    def _arrow2642(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("@id", Decode_uri)

    def _arrow2643(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("name", string)

    def _arrow2644(__unit: None=None) -> FSharpList[Source] | None:
        arg_15: Decoder_1[FSharpList[Source]] = list_1_3(ROCrate_decoder_3)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("derivesFrom", arg_15)

    return Sample(_arrow2642(), _arrow2643(), pattern_input[0], pattern_input[1], _arrow2644())


ROCrate_decoder: Decoder_1[Sample] = object(_arrow2645)

def ISAJson_encoder(id_map: Any | None, oa: Sample) -> IEncodable:
    def f(oa_1: Sample, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map_1(mapping, tupled_arg[1])

        def _arrow2649(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2648(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2648()

        def _arrow2651(value_2: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2650(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2650()

        def _arrow2652(oa_2: MaterialAttributeValue, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_1(id_map, oa_2)

        def _arrow2653(fv: FactorValue, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_2(id_map, fv)

        def _arrow2654(oa_3: Source, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_3(id_map, oa_3)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2649, ROCrate_genID(oa_1)), try_include("name", _arrow2651, oa_1.Name), try_include_list_opt("characteristics", _arrow2652, oa_1.Characteristics), try_include_list_opt("factorValues", _arrow2653, oa_1.FactorValues), try_include_list_opt("derivesFrom", _arrow2654, oa_1.DerivesFrom)]))
        class ObjectExpr2655(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr2655()

    if id_map is not None:
        def _arrow2656(s_1: Sample, id_map: Any=id_map, oa: Any=oa) -> str:
            return ROCrate_genID(s_1)

        return encode(_arrow2656, f, oa, id_map)

    else: 
        return f(oa)



ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "name", "characteristics", "factorValues", "derivesFrom", "@type", "@context"])

def _arrow2662(get: IGetters) -> Sample:
    def _arrow2657(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2658(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2659(__unit: None=None) -> FSharpList[MaterialAttributeValue] | None:
        arg_5: Decoder_1[FSharpList[MaterialAttributeValue]] = list_1_3(ISAJson_decoder_1)
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("characteristics", arg_5)

    def _arrow2660(__unit: None=None) -> FSharpList[FactorValue] | None:
        arg_7: Decoder_1[FSharpList[FactorValue]] = list_1_3(ISAJson_decoder_2)
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("factorValues", arg_7)

    def _arrow2661(__unit: None=None) -> FSharpList[Source] | None:
        arg_9: Decoder_1[FSharpList[Source]] = list_1_3(ISAJson_decoder_3)
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("derivesFrom", arg_9)

    return Sample(_arrow2657(), _arrow2658(), _arrow2659(), _arrow2660(), _arrow2661())


ISAJson_decoder: Decoder_1[Sample] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow2662)

__all__ = ["ROCrate_genID", "ROCrate_encoder", "ROCrate_additionalPropertyDecoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ....arctrl_core.ontology_annotation import OntologyAnnotation
from ....arctrl_core.Table.composite_cell import CompositeCell
from ....arctrl_core.template import Template
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.Table.templates import (Template_encoder, Template_decoder, Template_decoderCompressed, Template_encoderCompressed)
from ....fable_library.array_ import map
from ....fable_library.option import default_arg
from ....fable_library.result import FSharpResult_2
from ....fable_library.string_ import (to_fail, printf, to_text)
from ....fable_library.types import Array
from ....thoth_json_core.decode import array as array_2
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string
from .compression import (decode, encode)

__A_ = TypeVar("__A_")

def Templates_encoder(templates: Array[Template]) -> IEncodable:
    def mapping(template: Template, templates: Any=templates) -> IEncodable:
        return Template_encoder(template)

    values: Array[IEncodable] = map(mapping, templates, None)
    class ObjectExpr3848(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], templates: Any=templates) -> Any:
            def mapping_1(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map(mapping_1, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr3848()


Templates_decoder: Decoder_1[Array[Template]] = array_2(Template_decoder)

def Templates_fromJsonString(json_string: str) -> Array[Template]:
    try: 
        match_value: FSharpResult_2[Array[Template], str] = Decode_fromString(Templates_decoder, json_string)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as exn:
        return to_fail(printf("Error. Given json string cannot be parsed to Templates map: %A"))(exn)



def Templates_toJsonString(spaces: int, templates: Array[Template]) -> str:
    return to_string(spaces, Templates_encoder(templates))


def ARCtrl_Template__Template_fromJsonString_Static_Z721C83C5(json_string: str) -> Template:
    try: 
        match_value: FSharpResult_2[Template, str] = Decode_fromString(Template_decoder, json_string)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as exn:
        return to_fail(printf("Error. Given json string cannot be parsed to Template: %A"))(exn)



def ARCtrl_Template__Template_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[Template], str]:
    def _arrow3849(template: Template, spaces: Any=spaces) -> str:
        return to_string(default_spaces(spaces), Template_encoder(template))

    return _arrow3849


def ARCtrl_Template__Template_toJsonString_71136F3F(this: Template, spaces: int | None=None) -> str:
    return ARCtrl_Template__Template_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_Template__Template_fromCompressedJsonString_Static_Z721C83C5(s: str) -> Template:
    try: 
        match_value: FSharpResult_2[Template, str] = Decode_fromString(decode(Template_decoderCompressed), s)
        if match_value.tag == 1:
            raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

        else: 
            return match_value.fields[0]


    except Exception as e_1:
        arg_1: str = str(e_1)
        return to_fail(printf("Error. Unable to parse json string to ArcStudy: %s"))(arg_1)



def ARCtrl_Template__Template_toCompressedJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[Template], str]:
    def _arrow3850(obj: Template, spaces: Any=spaces) -> str:
        return to_string(default_arg(spaces, 0), encode(Template_encoderCompressed, obj))

    return _arrow3850


def ARCtrl_Template__Template_toCompressedJsonString_71136F3F(this: Template, spaces: int | None=None) -> str:
    return ARCtrl_Template__Template_toCompressedJsonString_Static_71136F3F(spaces)(this)


__all__ = ["Templates_encoder", "Templates_decoder", "Templates_fromJsonString", "Templates_toJsonString", "ARCtrl_Template__Template_fromJsonString_Static_Z721C83C5", "ARCtrl_Template__Template_toJsonString_Static_71136F3F", "ARCtrl_Template__Template_toJsonString_71136F3F", "ARCtrl_Template__Template_fromCompressedJsonString_Static_Z721C83C5", "ARCtrl_Template__Template_toCompressedJsonString_Static_71136F3F", "ARCtrl_Template__Template_toCompressedJsonString_71136F3F"]




from __future__ import annotations

import re
from datetime import datetime, timedelta, timezone
from math import fmod
from typing import Any

from .singleton_local_time_zone import local_time_zone
from .time_span import TimeSpan, total_microseconds
from .time_span import create as create_time_span
from .types import FSharpRef
from .util import DateKind


# Some of the code in this file has been adapted from
# https://github.com/microsoft/referencesource/blob/51cf7850defa8a17d815b4700b67116e3fa283c2/mscorlib/system/globalization/datetimeformat.cs

short_days = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]

long_days = ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]

short_months = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]

long_months = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]


def subtract(x: datetime, y: datetime | TimeSpan) -> datetime | TimeSpan:
    if isinstance(y, datetime):
        delta = x - y
        # ts.microseconds only contains the microseconds provided to the constructor
        # so we need to calculate the total microseconds ourselves
        delta_microseconds = delta.days * (24 * 3600 * 10**6) + delta.seconds * 10**6 + delta.microseconds
        return create_time_span(0, 0, 0, 0, 0, delta_microseconds)

    return x - timedelta(microseconds=total_microseconds(y))


def op_subtraction(x: datetime, y: datetime | TimeSpan) -> datetime | TimeSpan:
    return subtract(x, y)


def create(
    year: int,
    month: int,
    day: int,
    h: int = 0,
    m: int = 0,
    s: int = 0,
    ms: int = 0,
    mc: int = 0,
    kind: DateKind | None = None,
) -> datetime:
    if kind == DateKind.UTC:
        date = datetime(
            year=year,
            month=month,
            day=day,
            hour=h,
            minute=m,
            second=s,
            microsecond=mc + ms * 1_000,
            tzinfo=timezone.utc,
        )
    else:
        date = datetime(year, month, day, h, m, s, mc + ms * 1_000)
        if kind == DateKind.Local:
            date = date.astimezone(local_time_zone)

    return date


def year(d: datetime) -> int:
    return d.year


def month(d: datetime) -> int:
    return d.month


def day(d: datetime) -> int:
    return d.day


def hour(d: datetime) -> int:
    return d.hour


def minute(d: datetime) -> int:
    return d.minute


def second(d: datetime) -> int:
    return d.second


def millisecond(d: datetime) -> int:
    return d.microsecond // 1_000


def microsecond(d: datetime) -> int:
    return d.microsecond


def to_universal_time(d: datetime) -> datetime:
    return d.astimezone(timezone.utc)


def day_of_week(d: datetime) -> int:
    return (d.weekday() + 1) % 7


def day_of_year(d: datetime) -> int:
    return d.timetuple().tm_yday


def to_short_date_string(date: datetime) -> str:
    return datetime.strftime(date, "%m/%d/%Y")


def to_long_date_string(date: datetime) -> str:
    return datetime.strftime(date, "%A, %d %B %Y")


def to_short_time_string(date: datetime) -> str:
    return datetime.strftime(date, "%H:%M")


def to_long_time_string(date: datetime) -> str:
    return datetime.strftime(date, "%H:%M:%S")


def parse_repeat_token(format: str, pos: int, pattern_char: str) -> int:
    token_length = 0
    internal_pos = pos
    while internal_pos < len(format) and format[internal_pos] == pattern_char:
        internal_pos += 1
        token_length += 1

    return token_length


# Get the next character at the index of 'pos' in the 'format' string.
# Return value of -1 means 'pos' is already at the end of the 'format' string.
# Otherwise, return value is the int value of the next character.
def parse_next_char(format: str, pos: int) -> int:
    if pos >= len(format) - 1:
        return -1

    return ord(format[pos + 1])


def parse_quoted_string(format: str, pos: int) -> tuple[str, int]:
    begin_pos = pos
    format_length = len(format)
    # Get the character used to quote the string
    quote_char = format[pos]

    result = ""
    found_quote = False

    while pos < format_length:
        pos += 1
        current_char = format[pos]
        if current_char == quote_char:
            found_quote = True
            break
        elif current_char == "\\":
            if pos < format_length:
                pos += 1
                result += format[pos]
            else:
                # This means that '\'is the last character in the format string.
                raise ValueError("Invalid string format")
        else:
            result += current_char

    if not found_quote:
        # We couldn't find the matching quote
        raise ValueError(f"Invalid string format could not find matching quote for {quote_char}")

    return (result, pos - begin_pos + 1)


def date_to_string_with_custom_format(date: datetime, format: str, utc: bool) -> str:
    cursor_pos = 0
    token_length = 0
    result = ""
    localized_date = date.astimezone(timezone.utc) if utc else date

    while cursor_pos < len(format):
        token = format[cursor_pos]

        match token:
            case "d":
                token_length = parse_repeat_token(format, cursor_pos, "d")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.day)
                    case 2:
                        result += localized_date.strftime("%d")
                    case 3:
                        result += short_days[day_of_week(localized_date)]
                    case 4:
                        result += long_days[day_of_week(localized_date)]
                    case _:
                        pass
            case "f":
                token_length = parse_repeat_token(format, cursor_pos, "f")
                cursor_pos += token_length
                match token_length:
                    case 1 | 2 | 3 | 4 | 5 | 6:
                        precision = 10 ** (6 - token_length)
                        result += str(localized_date.microsecond // precision).zfill(token_length)
                    # Python datetime only support precision up to the microsecond
                    # so we need to fill the rest with 0
                    case 7:
                        result += str(localized_date.microsecond).zfill(6).ljust(token_length, "0")
                    case _:
                        pass
            case "F":
                token_length = parse_repeat_token(format, cursor_pos, "F")
                cursor_pos += token_length
                match token_length:
                    case 1 | 2 | 3 | 4 | 5 | 6:
                        precision = 10 ** (6 - token_length)
                        value = localized_date.microsecond // precision
                        if value != 0:
                            result += str(value).zfill(token_length)
                    # Python datetime only support precision up to the microsecond
                    # so we can't go further
                    # We alse need to pad start with 0 if the value is not 0
                    case 7:
                        value = localized_date.microsecond
                        if value != 0:
                            result += str(value).zfill(6).rstrip("0")
                    case _:
                        pass
            case "g":
                token_length = parse_repeat_token(format, cursor_pos, "g")
                cursor_pos += token_length
                match token_length:
                    case 1 | 2:
                        result += "A.D."
                    case _:
                        pass

            case "h":
                token_length = parse_repeat_token(format, cursor_pos, "h")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.hour % 12)
                    case 2:
                        result += localized_date.strftime("%I")
                    case _:
                        pass
            case "H":
                token_length = parse_repeat_token(format, cursor_pos, "H")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.hour)
                    case 2:
                        result += localized_date.strftime("%H")
                    case _:
                        pass
            case "K":
                token_length = parse_repeat_token(format, cursor_pos, "K")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        match kind(date):
                            case DateKind.UTC:
                                result += "Z"
                            case DateKind.Local:
                                # %:z is not a perfect match for the .NET equivalent
                                # but it seems to do the job
                                # If needed we can probably compute the offset manually
                                # and format it ourselves
                                result += date.strftime("%:z")
                            case DateKind.Unspecified:
                                result += ""
                    case _:
                        pass
            case "m":
                token_length = parse_repeat_token(format, cursor_pos, "m")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.minute)
                    case 2:
                        result += localized_date.strftime("%M")
                    case _:
                        pass
            case "M":
                token_length = parse_repeat_token(format, cursor_pos, "M")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.month)
                    case 2:
                        result += localized_date.strftime("%m")
                    case 3:
                        result += short_months[month(localized_date) - 1]
                    case 4:
                        result += long_months[month(localized_date) - 1]
                    case _:
                        pass
            case "s":
                token_length = parse_repeat_token(format, cursor_pos, "s")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.second)
                    case 2:
                        result += localized_date.strftime("%S")
                    case _:
                        pass
            case "t":
                token_length = parse_repeat_token(format, cursor_pos, "t")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += localized_date.strftime("%p")[:1]
                    case 2:
                        result += localized_date.strftime("%p")
                    case _:
                        pass
            case "y":
                token_length = parse_repeat_token(format, cursor_pos, "y")
                cursor_pos += token_length
                match token_length:
                    case 1:
                        result += str(localized_date.year % 100)
                    case 2:
                        result += str(localized_date.year)[-2:].zfill(2)
                    case length:
                        result += str(localized_date.year).zfill(length)
            case "z":
                token_length = parse_repeat_token(format, cursor_pos, "z")
                cursor_pos += token_length

                match kind(date):
                    case DateKind.UTC:
                        utc_offet_text = localized_date.strftime("%z")
                    case DateKind.Local:
                        utc_offet_text = localized_date.strftime("%z")
                    case DateKind.Unspecified:
                        utc_offet_text = to_local_time(date).strftime("%z")

                sign = utc_offet_text[:1]
                hours = int(utc_offet_text[1:3])
                minutes = int(utc_offet_text[3:5])

                match token_length:
                    case 1:
                        result += f"{sign}{hours}"
                    case 2:
                        result += f"{sign}{hours:02}"
                    case 3 | _:
                        result += f"{sign}{hours:02}:{minutes:02}"
            case ":":
                cursor_pos += 1
                result += ":"
            case "/":
                cursor_pos += 1
                result += "/"
            case "'" | '"':
                quoted_string, quoted_string_length = parse_quoted_string(format, cursor_pos)
                cursor_pos += quoted_string_length
                result += quoted_string
            case "%":
                next_char = parse_next_char(format, cursor_pos)
                if next_char >= 0 and next_char != ord("%"):
                    cursor_pos += 2
                    result += date_to_string_with_custom_format(date, chr(next_char), utc)
                else:
                    raise Exception("Invalid string format")
            case "\\":
                next_char = parse_next_char(format, cursor_pos)
                if next_char >= 0:
                    result += chr(next_char)
                    cursor_pos += 2
                else:
                    raise Exception("Invalid string format")
            case _:
                cursor_pos += 1
                result += token
                pass

    return result


def date_to_string_with_offset(date: datetime, format: str | None = None) -> str:
    utc = date.tzinfo == timezone.utc

    match format:
        case None:
            raise NotImplementedError("date_to_string_with_offset")
        case "O" | "o":
            return date.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        case _ if len(format) == 1:
            raise Exception("Unrecognized Date print format")
        case _:
            return date_to_string_with_custom_format(date, format, utc)


def date_to_string_with_kind(date: datetime, format: str | None = None) -> str:
    utc = date.tzinfo == timezone.utc

    if not format:
        return date.isoformat() if utc else str(date)
    elif len(format) == 1:
        if format == "d":
            return to_short_date_string(date)
        elif format == "D":
            return to_long_date_string(date)
        elif format == "T":
            return to_long_time_string(date)
        elif format == "t":
            return to_short_time_string(date)
        elif format == "O" or format == "o":
            return date.astimezone().isoformat(timespec="milliseconds")
        else:
            raise Exception("Unrecognized Date print format")

    else:
        return date_to_string_with_custom_format(date, format, utc)


def to_string(date: datetime, format: str | None = None, provider: Any | None = None) -> str:
    if date.tzinfo:
        return date_to_string_with_offset(date, format)

    return date_to_string_with_kind(date, format)


def now() -> datetime:
    return datetime.now()


def utc_now() -> datetime:
    return datetime.utcnow()


def today() -> datetime:
    return datetime.replace(now(), hour=0, minute=0, second=0, microsecond=0)


def to_local_time(date: datetime) -> datetime:
    return date.astimezone().replace(tzinfo=local_time_zone)


def compare(x: datetime, y: datetime) -> int:
    if x == y:
        return 0

    if x < y:
        return -1

    return 1


def equals(x: datetime, y: datetime) -> bool:
    return x == y


def max_value() -> datetime:
    return datetime.max


def min_value() -> datetime:
    return datetime.min


def op_addition(x: datetime, y: TimeSpan) -> datetime:
    return x + timedelta(microseconds=total_microseconds(y))


def add(x: datetime, y: TimeSpan) -> datetime:
    return op_addition(x, y)


def parse(string: str) -> datetime:
    try:
        return datetime.fromisoformat(string).astimezone()
    except ValueError:
        pass

    # For the time-only formats, needs a special treatment
    # because in Python, they are set in 1900-01-01 while
    # in F# they are set in the current date
    hoursFormats = {
        # Matches a time string in 24-hour format "HH:MM:SS"
        r"^\d{1,2}:\d{1,2}:\d{2}$": "%H:%M:%S",
        # Matches a time string in 12-hour format with AM/PM "HH:MM:SS AM" or "HH:MM:SS PM"
        r"^(0?[1-9]|1[0-2]):([0-5][1-9]|0?[0-9]):([0-5][0-9]|0?[0-9]) [AP]M$": "%I:%M:%S %p",
        r"^\d{1,2}:\d{1,2}:\d{2} [AP]M$": "%H:%M:%S %p",
    }

    for pattern, format in hoursFormats.items():
        if re.fullmatch(pattern, string):
            hourDate = datetime.strptime(string, format)

            # If the hour is 0 PM, then in .NET it is 12 PM (python keeps it as 0)
            hourOffset = 12 if hourDate.hour == 0 and string.endswith(" PM") else 0

            return datetime.replace(
                now(),
                hour=hourDate.hour + hourOffset,
                minute=hourDate.minute,
                second=hourDate.second,
                microsecond=0,
            )

    formats = {
        # 9/10/2014 1:50:34 PM
        r"^(0?[1-9]|1[0-2])\/(0?[1-9]|1[0-2])\/\d{4} ([0-9]|(0|1)[0-9]|2[0-4]):([0-5][0-9]|0?[0-9]):([0-5][0-9]|0?[0-9]) [AP]M$": "%m/%d/%Y %I:%M:%S %p",
        # 9/10/2014 1:50:34
        r"^(0?[1-9]|1[0-2])\/(0?[1-9]|1[0-2])\/\d{4} ([0-9]|(0|1)[0-9]|2[0-4]):([0-5][0-9]|0?[0-9]):([0-5][0-9]|0?[0-9])$": "%m/%d/%Y %H:%M:%S",
    }

    for pattern, format in formats.items():
        if re.fullmatch(pattern, string):
            return datetime.strptime(string, format)

    # Matches a datetime string in the format "YYYY-MM-DDTHH:MM:SS.ffffff". This format usually parses with
    # `fromisoformat`, but not with Python 3.10
    iso_format_regex = r"(?P<header>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.)(?P<microseconds>\d{0,7})"

    # In Python %f only supports exactly 6 digits so we need to adapt the string
    # If microseconds has 7 digits, we remove the last one
    # If microseconds has less than 6 digits, we pad it with 0
    def adapt_microseconds(m: re.Match[str]) -> str:
        microseconds = m.group("microseconds")
        header_text = m.group("header")

        if len(microseconds) == 7:
            microseconds = microseconds[:-1]

        return header_text + microseconds.ljust(6, "0")

    if re.fullmatch(iso_format_regex, string):
        adapted_string = re.sub(iso_format_regex, adapt_microseconds, string)
        return datetime.strptime(adapted_string, "%Y-%m-%dT%H:%M:%S.%f")

    raise ValueError("Unsupported format by Fable: %s" % (string))


def try_parse(string: str, def_value: FSharpRef[datetime]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


def date(d: datetime) -> datetime:
    return create(d.year, d.month, d.day, 0, 0, 0, 0, 0, kind(d))


def is_leap_year(year: int) -> bool:
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def days_in_month(year: int, month: int) -> int:
    if month == 2:
        return 29 if is_leap_year(year) else 28

    if month in (4, 6, 9, 11):
        return 30

    return 31


def add_years(d: datetime, v: int) -> datetime:
    new_month = month(d)
    new_year = year(d) + v
    _days_in_month = days_in_month(new_year, new_month)
    new_day = min(_days_in_month, day(d))
    return create(new_year, new_month, new_day, hour(d), minute(d), second(d), millisecond(d), microsecond(d), kind(d))


def add_months(d: datetime, v: int) -> datetime:
    new_month = month(d) + v
    new_month_ = 0
    year_offset = 0
    if new_month > 12:
        new_month_ = int(fmod(new_month, 12))
        year_offset = new_month // 12
        new_month = new_month_
    elif new_month < 1:
        new_month_ = 12 + int(fmod(new_month, 12))
        year_offset = new_month // 12 + (-1 if new_month_ == 12 else 0)
        new_month = new_month_
    new_year = year(d) + year_offset
    _days_in_month = days_in_month(new_year, new_month)
    new_day = min(_days_in_month, day(d))
    return create(new_year, new_month, new_day, hour(d), minute(d), second(d), millisecond(d), microsecond(d), kind(d))


def add_days(d: datetime, v: int) -> datetime:
    return d + timedelta(days=v)


def add_hours(d: datetime, v: int) -> datetime:
    return d + timedelta(hours=v)


def add_minutes(d: datetime, v: int) -> datetime:
    return d + timedelta(minutes=v)


def add_seconds(d: datetime, v: int) -> datetime:
    return d + timedelta(seconds=v)


def add_milliseconds(d: datetime, v: int) -> datetime:
    return d + timedelta(milliseconds=v)


def add_microseconds(d: datetime, v: int) -> datetime:
    return d + timedelta(microseconds=v)


def kind(d: datetime) -> DateKind:
    if d.tzinfo == timezone.utc:
        return DateKind.UTC

    elif d.tzinfo is None:
        return DateKind.Unspecified

    # Should we check that tzinfo is local_time_zone?
    return DateKind.Local


def specify_kind(d: datetime, kind: DateKind) -> datetime:
    return create(year(d), month(d), day(d), hour(d), minute(d), second(d), millisecond(d), microsecond(d), kind)


def ticks(d: datetime) -> int:
    # Note: It can happens that Ticks differs a little bit from the .NET implementation
    # because of some rounding/precision issues in Python
    # DateTime(1, 1, 1, 0, 0, 0, 0, 99, DateTimeKind.Utc).Ticks should be 990
    # but Python returns 1040
    # because d.timestamp():
    # - returns     -62135596799.9999
    # - instead of  -62135596800000
    # compute timestamp in microseconds
    return unix_epoch_microseconds_to_ticks(int(d.timestamp() * 1_000_000), date_offset(d) * 1_000)


def unix_epoch_microseconds_to_ticks(us: int, offset: int) -> int:
    return int(((us + 62135596800000000) + offset) * 10)


def ticks_to_unix_epoch_microseconds(ticks: int) -> int:
    return int((ticks - 621355968000000000) // 10)


def date_offset(d: datetime) -> int:
    if d.tzinfo == timezone.utc:
        return 0
    else:
        utc_offset = d.utcoffset()

        # Is it correct to force an offset to local time
        # for DateKind.Unspecified?
        if utc_offset is None:
            forced_utc_offset = d.astimezone().utcoffset()
            assert forced_utc_offset is not None
            return int(forced_utc_offset.total_seconds() * 1_000)
        else:
            return int(utc_offset.total_seconds() * 1_000)

    # return 0 if d.tzinfo == timezone.utc else


def create_from_epoch_microseconds(us: int, kind: DateKind | None = None) -> datetime:
    if kind == DateKind.UTC:
        date = datetime.fromtimestamp(us / 1_000_000, timezone.utc)
    else:
        date = datetime.fromtimestamp(us / 1_000_000)
        if kind == DateKind.Local:
            date = date.astimezone()

    return date


def from_ticks(ticks: int, kind: DateKind | None = None) -> datetime:
    # Better default than Unspecified
    kind = kind or DateKind.Local
    date = create_from_epoch_microseconds(ticks_to_unix_epoch_microseconds(ticks), kind)

    # Ticks are local to offset (in this case, either UTC or Local/Unknown).
    # If kind is anything but UTC, that means that the tick number was not
    # in utc, thus getTime() cannot return UTC, and needs to be shifted.
    if kind != DateKind.UTC:
        date = create_from_epoch_microseconds(int(date.timestamp() * 1_000_000 - date_offset(date) * 1_000), kind)

    return date


__all__ = [
    "subtract",
    "op_subtraction",
    "create",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "to_universal_time",
    "day_of_week",
    "day_of_year",
    "to_short_date_string",
    "to_long_date_string",
    "to_short_time_string",
    "to_long_time_string",
    "to_string",
    "now",
    "utc_now",
    "today",
    "to_local_time",
    "compare",
    "equals",
    "max_value",
    "min_value",
    "op_addition",
    "add",
    "parse",
    "try_parse",
    "date",
    "is_leap_year",
    "days_in_month",
    "add_years",
    "add_months",
    "add_days",
    "add_hours",
    "add_minutes",
    "add_seconds",
    "add_milliseconds",
    "add_microseconds",
    "kind",
    "specify_kind",
    "ticks",
    "date_offset",
    "from_ticks",
]



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar, cast


_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")
_W = TypeVar("_W")


class Some(Generic[_T]):
    ___slots__ = "value"

    def __init__(self, value: _T):
        self.value = value

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        if other is None:
            return False

        if self.value == other.value:
            return True

        return False

    def __str__(self):
        return f"Some {self.value}"

    def __repr__(self):
        return str(self)


def default_arg(opt: _T | None, default_value: _T) -> _T:
    return value(opt) if opt is not None else default_value


def default_arg_with(opt: _T | None, def_thunk: Callable[[], _T]) -> _T:
    return value(opt) if opt is not None else def_thunk()


def filter(predicate: Callable[[_T], bool], opt: _T | None) -> _T | None:
    if opt is not None:
        return opt if predicate(value(opt)) else None
    return opt


def map(mapping: Callable[[_T], _U], opt: _T | None) -> _U | None:
    return some(mapping(value(opt))) if opt is not None else None


def map2(mapping: Callable[[_T, _U], _V], opt1: _T | None, opt2: _U | None) -> _V | None:
    return mapping(value(opt1), value(opt2)) if (opt1 is not None and opt2 is not None) else None


def map3(
    mapping: Callable[[_T, _U, _V], _W],
    opt1: _T | None,
    opt2: _U | None,
    opt3: _V | None,
) -> _W | None:
    return (
        mapping(value(opt1), value(opt2), value(opt3))
        if (opt1 is not None and opt2 is not None and opt3 is not None)
        else None
    )


def some(x: Any) -> Any | None:
    return Some[Any](x) if x is None or isinstance(x, Some) else x


def value(x: _T | None) -> _T:
    if x is None:
        raise Exception("Option has no value")
    elif isinstance(x, Some):
        y = cast(Some[_T], x)
        return y.value
    return x


def of_nullable(x: _T | None) -> _T | None:
    return x


def to_nullable(x: _T | None) -> _T | None:
    return None if x is None else value(x)


def flatten(x: _T | None | None) -> _T | None:
    return x if x is None else value(x)


def to_array(opt: _T | None) -> list[_T]:
    return [] if opt is None else [value(opt)]


def bind(binder: Callable[[_T], _U | None], opt: _T | None) -> _U | None:
    return binder(value(opt)) if opt is not None else None


def or_else(opt: _T | None, if_none: _T | None) -> _T | None:
    return if_none if opt is None else opt


def or_else_with(opt: _T | None, if_none_thunk: Callable[[], _T | None]) -> _T | None:
    return if_none_thunk() if opt is None else opt


__all__ = [
    "bind",
    "default_arg",
    "default_arg_with",
    "flatten",
    "map",
    "map2",
    "map3",
    "of_nullable",
    "some",
    "Some",
    "to_array",
    "to_nullable",
    "value",
]



from __future__ import annotations
from typing import (Any, TypeVar)
from ..fable_library.types import FSharpRef
from ..fable_library.array_ import (fill, map)
from ..fable_library.map_util import (try_get_value, add_to_dict)
from ..fable_library.result import FSharpResult_2
from ..fable_library.seq import iterate
from ..fable_library.types import Array
from ..thoth_json_core.decode import (array as array_2, string, int_1)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1, ErrorReason_1, IDecoderHelpers_1)

__A_ = TypeVar("__A_")

def array_from_map(otm: Any) -> Array[str]:
    a: Array[str] = fill([0] * len(otm), 0, len(otm), "")
    def action(kv: Any, otm: Any=otm) -> None:
        a[kv[1]] = kv[0]

    iterate(action, otm)
    return a


def encoder(ot: Array[str]) -> IEncodable:
    def _arrow1960(value: str, ot: Any=ot) -> IEncodable:
        class ObjectExpr1959(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1959()

    values: Array[IEncodable] = map(_arrow1960, ot, None)
    class ObjectExpr1961(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], ot: Any=ot) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers_1)

            arg: Array[__A_] = map(mapping, values, None)
            return helpers_1.encode_array(arg)

    return ObjectExpr1961()


decoder: Decoder_1[Array[str]] = array_2(string)

def encode_string(otm: Any, s: str) -> IEncodable:
    match_value: int | None
    pattern_input: tuple[bool, int]
    out_arg: int = None or 0
    def _arrow1962(__unit: None=None, otm: Any=otm, s: Any=s) -> int:
        return out_arg

    def _arrow1963(v: int, otm: Any=otm, s: Any=s) -> None:
        nonlocal out_arg
        out_arg = v or 0

    pattern_input = (try_get_value(otm, s, FSharpRef(_arrow1962, _arrow1963)), out_arg)
    match_value = pattern_input[1] if pattern_input[0] else None
    if match_value is None:
        i_1: int = len(otm) or 0
        add_to_dict(otm, s, i_1)
        class ObjectExpr1964(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], otm: Any=otm, s: Any=s) -> Any:
                return helpers_1.encode_signed_integral_number(i_1)

        return ObjectExpr1964()

    else: 
        i: int = match_value or 0
        class ObjectExpr1965(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], otm: Any=otm, s: Any=s) -> Any:
                return helpers.encode_signed_integral_number(i)

        return ObjectExpr1965()



def decode_string(ot: Array[str]) -> Decoder_1[str]:
    class ObjectExpr1966(Decoder_1[str]):
        def Decode(self, s: IDecoderHelpers_1[Any], json: Any, ot: Any=ot) -> FSharpResult_2[str, tuple[str, ErrorReason_1[__A_]]]:
            match_value: FSharpResult_2[int, tuple[str, ErrorReason_1[__A_]]] = int_1.Decode(s, json)
            return FSharpResult_2(1, match_value.fields[0]) if (match_value.tag == 1) else FSharpResult_2(0, ot[match_value.fields[0]])

    return ObjectExpr1966()


__all__ = ["array_from_map", "encoder", "decoder", "encode_string", "decode_string"]




from __future__ import annotations
from ..dynamic_obj.dynamic_obj import DynamicObj
from ..dynamic_obj.dyn_obj import set_optional_property
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import FSharpRef
from .ldcontext import LDContext
from .ldobject import (LDNode, LDNode_reflection)

def _expr1582() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.ArcROCrateMetadata", None, ArcROCrateMetadata, LDNode_reflection())


class ArcROCrateMetadata(LDNode):
    def __init__(self, about: LDNode | None=None) -> None:
        super().__init__("ro-crate-metadata", ["CreativeWork"])
        this: FSharpRef[ArcROCrateMetadata] = FSharpRef(None)
        this.contents = self
        self.init_00405: int = 1
        set_optional_property("about", about, this.contents)
        conforms_to: DynamicObj = DynamicObj()
        conforms_to.SetProperty("@id", "https://w3id.org/ro/crate/1.1")
        this.contents.SetProperty("conformsTo", conforms_to)
        context: LDContext = LDContext()
        context.AddMapping("sdo", "http://schema.org/")
        context.AddMapping("arc", "http://purl.org/nfdi4plants/ontology/")
        context.AddMapping("CreativeWork", "sdo:CreativeWork")
        context.AddMapping("about", "sdo:about")
        context.AddMapping("conformsTo", "sdo:conformsTo")
        this.contents.SetProperty("@context", context)


ArcROCrateMetadata_reflection = _expr1582

def ArcROCrateMetadata__ctor_Z7020538(about: LDNode | None=None) -> ArcROCrateMetadata:
    return ArcROCrateMetadata(about)


__all__ = ["ArcROCrateMetadata_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ...fable_library.list import (length, empty, FSharpList, choose)
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, string_type, option_type, list_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ..ontology_annotation import OntologyAnnotation
from .material_attribute_value import (MaterialAttributeValue, MaterialAttributeValue_reflection)
from .material_type import (MaterialType as MaterialType_1, MaterialType__get_AsString, MaterialType_reflection)

def _expr758() -> TypeInfo:
    return record_type("ARCtrl.Process.Material", [], Material, lambda: [("ID", option_type(string_type)), ("Name", option_type(string_type)), ("MaterialType", option_type(MaterialType_reflection())), ("Characteristics", option_type(list_type(MaterialAttributeValue_reflection()))), ("DerivesFrom", option_type(list_type(Material_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Material(Record):
    ID: str | None
    Name: str | None
    MaterialType: MaterialType_1 | None
    Characteristics: FSharpList[MaterialAttributeValue] | None
    DerivesFrom: FSharpList[Material] | None
    def Print(self, __unit: None=None) -> str:
        this: Material = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Material = self
        chars: int = length(default_arg(this.Characteristics, empty())) or 0
        match_value: MaterialType_1 | None = this.MaterialType
        if match_value is None:
            arg_3: str = Material__get_NameText(this)
            return to_text(printf("%s [%i characteristics]"))(arg_3)(chars)

        else: 
            t: MaterialType_1 = match_value
            arg: str = Material__get_NameText(this)
            arg_1: str = MaterialType__get_AsString(t)
            return to_text(printf("%s [%s; %i characteristics]"))(arg)(arg_1)(chars)



Material_reflection = _expr758

def Material_make(id: str | None=None, name: str | None=None, material_type: MaterialType_1 | None=None, characteristics: FSharpList[MaterialAttributeValue] | None=None, derives_from: FSharpList[Material] | None=None) -> Material:
    return Material(id, name, material_type, characteristics, derives_from)


def Material_create_Z66909A6D(Id: str | None=None, Name: str | None=None, MaterialType: MaterialType_1 | None=None, Characteristics: FSharpList[MaterialAttributeValue] | None=None, DerivesFrom: FSharpList[Material] | None=None) -> Material:
    return Material_make(Id, Name, MaterialType, Characteristics, DerivesFrom)


def Material_get_empty(__unit: None=None) -> Material:
    return Material_create_Z66909A6D()


def Material__get_NameText(this: Material) -> str:
    return default_arg(this.Name, "")


def Material_getUnits_ZBCDEB61(m: Material) -> FSharpList[OntologyAnnotation]:
    def chooser(c: MaterialAttributeValue, m: Any=m) -> OntologyAnnotation | None:
        return c.Unit

    return choose(chooser, default_arg(m.Characteristics, empty()))


def Material_setCharacteristicValues(values: FSharpList[MaterialAttributeValue], m: Material) -> Material:
    return Material(m.ID, m.Name, m.MaterialType, values, m.DerivesFrom)


__all__ = ["Material_reflection", "Material_make", "Material_create_Z66909A6D", "Material_get_empty", "Material__get_NameText", "Material_getUnits_ZBCDEB61", "Material_setCharacteristicValues"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (is_empty, head, cons, tail as tail_25, singleton, reverse, empty, of_array_with_tail, FSharpList, iterate, iterate_indexed, map, transpose, choose)
from ...fable_library.seq import max
from ...fable_library.set import (of_seq, FSharpSet__Contains, add)
from ...fable_library.string_ import (to_fail, printf, to_text)
from ...fable_library.types import to_string
from ...fable_library.util import (compare_primitives, ignore)
from ..Cells.fs_cell import (FsCell, DataType)
from ..Cells.fs_cells_collection import FsCellsCollection
from ..fs_address import FsAddress
from ..fs_row import FsRow
from ..fs_workbook import FsWorkbook
from ..fs_worksheet import FsWorksheet
from ..Ranges.fs_range_address import FsRangeAddress
from ..Tables.fs_table import FsTable
from ..Tables.fs_table_field import FsTableField
from .types import (SheetElement, TableElement__get_IsColumn, ColumnElement, TableElement, RowElement, ColumnIndex__get_Index, RowIndex__get_Index, RowIndex, WorkbookElement, Workbook)

def split_rows_and_columns(els: FSharpList[SheetElement]) -> FSharpList[tuple[str, FSharpList[SheetElement]]]:
    def loop(in_rows_mut: bool, in_columns_mut: bool, current_mut: FSharpList[SheetElement], remaining_mut: FSharpList[SheetElement], agg_mut: FSharpList[tuple[str, FSharpList[SheetElement]]], els: Any=els) -> FSharpList[tuple[str, FSharpList[SheetElement]]]:
        while True:
            (in_rows, in_columns, current, remaining, agg) = (in_rows_mut, in_columns_mut, current_mut, remaining_mut, agg_mut)
            if not is_empty(remaining):
                if head(remaining).tag == 4:
                    if in_columns:
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = cons(SheetElement(4, head(remaining).fields[0]), current)
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue

                    elif in_rows:
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = singleton(SheetElement(4, head(remaining).fields[0]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = cons(("Rows", reverse(current)), agg)
                        continue

                    else: 
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = singleton(SheetElement(4, head(remaining).fields[0]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue


                elif head(remaining).tag == 3:
                    if in_columns:
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = cons(SheetElement(3, head(remaining).fields[0], head(remaining).fields[1]), current)
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue

                    elif in_rows:
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = singleton(SheetElement(3, head(remaining).fields[0], head(remaining).fields[1]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = cons(("Rows", reverse(current)), agg)
                        continue

                    else: 
                        in_rows_mut = False
                        in_columns_mut = True
                        current_mut = singleton(SheetElement(3, head(remaining).fields[0], head(remaining).fields[1]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue


                elif head(remaining).tag == 2:
                    if in_rows:
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = cons(SheetElement(2, head(remaining).fields[0]), current)
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue

                    elif in_columns:
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = singleton(SheetElement(2, head(remaining).fields[0]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = cons(("Columns", reverse(current)), agg)
                        continue

                    else: 
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = singleton(SheetElement(2, head(remaining).fields[0]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue


                elif head(remaining).tag == 1:
                    if in_rows:
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = cons(SheetElement(1, head(remaining).fields[0], head(remaining).fields[1]), current)
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue

                    elif in_columns:
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = singleton(SheetElement(1, head(remaining).fields[0], head(remaining).fields[1]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = cons(("Columns", reverse(current)), agg)
                        continue

                    else: 
                        in_rows_mut = True
                        in_columns_mut = False
                        current_mut = singleton(SheetElement(1, head(remaining).fields[0], head(remaining).fields[1]))
                        remaining_mut = tail_25(remaining)
                        agg_mut = agg
                        continue


                elif head(remaining).tag == 0:
                    if in_rows:
                        in_rows_mut = False
                        in_columns_mut = False
                        current_mut = empty()
                        remaining_mut = tail_25(remaining)
                        agg_mut = of_array_with_tail([("Table", singleton(SheetElement(0, head(remaining).fields[0], head(remaining).fields[1]))), ("Rows", reverse(current))], agg)
                        continue

                    elif in_columns:
                        in_rows_mut = False
                        in_columns_mut = False
                        current_mut = empty()
                        remaining_mut = tail_25(remaining)
                        agg_mut = of_array_with_tail([("Table", singleton(SheetElement(0, head(remaining).fields[0], head(remaining).fields[1]))), ("Columns", reverse(current))], agg)
                        continue

                    else: 
                        in_rows_mut = False
                        in_columns_mut = False
                        current_mut = empty()
                        remaining_mut = tail_25(remaining)
                        agg_mut = cons(("Table", singleton(SheetElement(0, head(remaining).fields[0], head(remaining).fields[1]))), agg)
                        continue


                else: 
                    raise Exception("Unknown element combination when grouping Sheet elements")


            elif in_rows:
                return cons(("Rows", reverse(current)), agg)

            elif in_columns:
                return cons(("Columns", reverse(current)), agg)

            else: 
                return agg

            break

    return reverse(loop(False, False, empty(), els, empty()))


def FsSpreadsheet_DSL_Workbook__Workbook_parseTable_Static(cell_collection: FsCellsCollection, table: FsTable, els: FSharpList[TableElement]) -> None:
    def action_1(col_2: FSharpList[tuple[DataType, Any]], cell_collection: Any=cell_collection, table: Any=table, els: Any=els) -> None:
        if not is_empty(col_2):
            field: FsTableField = table.Field(to_string(head(col_2)[1]), cell_collection)
            def action(i: int, tupled_arg: tuple[DataType, Any], col_2: Any=col_2) -> None:
                cell_4: FsCell = field.Column.Cell(i + 2, cell_collection)
                cell_4.DataType = tupled_arg[0]
                cell_4.Value = tupled_arg[1]

            iterate_indexed(action, tail_25(col_2))

        else: 
            raise Exception("Empty column")


    def mapping_1(col: TableElement, cell_collection: Any=cell_collection, table: Any=table, els: Any=els) -> FSharpList[tuple[DataType, Any]]:
        if col.tag == 1:
            def mapping(cell: ColumnElement, col: Any=col) -> tuple[DataType, Any]:
                if cell.tag == 1:
                    return cell.fields[0]

                else: 
                    raise Exception("Indexed cells not supported in column transformation")


            return map(mapping, col.fields[0])

        else: 
            raise Exception("Indexed columns not supported in table transformation")


    def mapping_3(row: TableElement, cell_collection: Any=cell_collection, table: Any=table, els: Any=els) -> FSharpList[tuple[DataType, Any]]:
        if row.tag == 0:
            def mapping_2(cell_2: RowElement, row: Any=row) -> tuple[DataType, Any]:
                if cell_2.tag == 1:
                    return cell_2.fields[0]

                else: 
                    raise Exception("Indexed cells not supported in row transformation")


            return map(mapping_2, row.fields[0])

        else: 
            raise Exception("Indexed rows not supported in table transformation")


    iterate(action_1, map(mapping_1, els) if TableElement__get_IsColumn(head(els)) else transpose(map(mapping_3, els)))


def FsSpreadsheet_DSL_Workbook__Workbook_parseRow_Static(cell_collection: FsCellsCollection, row: FsRow, els: FSharpList[RowElement]) -> None:
    def chooser(el: RowElement, cell_collection: Any=cell_collection, row: Any=row, els: Any=els) -> int | None:
        if el.tag == 0:
            return ColumnIndex__get_Index(el.fields[0])

        else: 
            return None


    class ObjectExpr299:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    cell_index_set: Any = of_seq(choose(chooser, els), ObjectExpr299())
    def action(el_1: RowElement, cell_collection: Any=cell_collection, row: Any=row, els: Any=els) -> None:
        if el_1.tag == 1:
            def _arrow300(__unit: None=None, el_1: Any=el_1) -> int:
                nonlocal cell_index_set
                i_1: int = 1
                while FSharpSet__Contains(cell_index_set, i_1):
                    i_1 = (i_1 + 1) or 0
                cell_index_set = add(i_1, cell_index_set)
                return i_1

            cell_1: FsCell = row.Item(_arrow300())
            cell_1.DataType = el_1.fields[0][0]
            cell_1.Value = el_1.fields[0][1]

        else: 
            cell: FsCell = row.Item(ColumnIndex__get_Index(el_1.fields[0]))
            cell.DataType = el_1.fields[1][0]
            cell.Value = el_1.fields[1][1]


    iterate(action, els)


def FsSpreadsheet_DSL_Workbook__Workbook_parseSheet_Static(sheet: FsWorksheet, els: FSharpList[SheetElement]) -> None:
    def chooser(el: SheetElement, sheet: Any=sheet, els: Any=els) -> int | None:
        if el.tag == 1:
            return RowIndex__get_Index(el.fields[0])

        else: 
            return None


    class ObjectExpr301:
        @property
        def Compare(self) -> Callable[[int, int], int]:
            return compare_primitives

    row_index_set: Any = add(0, of_seq(choose(chooser, els), ObjectExpr301()))
    def action_3(_arg: tuple[str, FSharpList[SheetElement]], sheet: Any=sheet, els: Any=els) -> None:
        (pattern_matching_result, l, name, table_elements, l_1, s) = (None, None, None, None, None, None)
        if _arg[0] == "Columns":
            pattern_matching_result = 0
            l = _arg[1]

        elif _arg[0] == "Table":
            if not is_empty(_arg[1]):
                if head(_arg[1]).tag == 0:
                    if is_empty(tail_25(_arg[1])):
                        pattern_matching_result = 1
                        name = head(_arg[1]).fields[0]
                        table_elements = head(_arg[1]).fields[1]

                    else: 
                        pattern_matching_result = 3
                        s = _arg[0]


                else: 
                    pattern_matching_result = 3
                    s = _arg[0]


            else: 
                pattern_matching_result = 3
                s = _arg[0]


        elif _arg[0] == "Rows":
            pattern_matching_result = 2
            l_1 = _arg[1]

        else: 
            pattern_matching_result = 3
            s = _arg[0]

        if pattern_matching_result == 0:
            columns: FSharpList[SheetElement] = l
            class ObjectExpr302:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            base_row_index: int = (1 + max(row_index_set, ObjectExpr302())) or 0
            def chooser_1(col: SheetElement, _arg: Any=_arg) -> int | None:
                if col.tag == 3:
                    return ColumnIndex__get_Index(col.fields[0])

                else: 
                    return None


            class ObjectExpr303:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            column_index_set: Any = of_seq(choose(chooser_1, columns), ObjectExpr303())
            def action_1(col_1: SheetElement, _arg: Any=_arg) -> None:
                pattern_input: tuple[int, FSharpList[ColumnElement]]
                if col_1.tag == 3:
                    pattern_input = (ColumnIndex__get_Index(col_1.fields[0]), col_1.fields[1])

                elif col_1.tag == 4:
                    def _arrow304(__unit: None=None, col_1: Any=col_1) -> int:
                        nonlocal column_index_set
                        i_3: int = 1
                        while FSharpSet__Contains(column_index_set, i_3):
                            i_3 = (i_3 + 1) or 0
                        column_index_set = add(i_3, column_index_set)
                        return i_3

                    pattern_input = (_arrow304(), col_1.fields[0])

                else: 
                    raise Exception("Expected column elements")

                elements_2: FSharpList[ColumnElement] = pattern_input[1]
                col_i: int = pattern_input[0] or 0
                def chooser_2(el_1: ColumnElement, col_1: Any=col_1) -> int | None:
                    if el_1.tag == 0:
                        return RowIndex__get_Index(el_1.fields[0])

                    else: 
                        return None


                class ObjectExpr305:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                cell_index_set: Any = of_seq(choose(chooser_2, elements_2), ObjectExpr305())
                def action(el_2: ColumnElement, col_1: Any=col_1) -> None:
                    nonlocal row_index_set
                    if el_2.tag == 1:
                        def _arrow306(__unit: None=None, el_2: Any=el_2) -> int:
                            nonlocal cell_index_set
                            i_6: int = 1
                            while FSharpSet__Contains(cell_index_set, i_6):
                                i_6 = (i_6 + 1) or 0
                            cell_index_set = add(i_6, cell_index_set)
                            return i_6

                        row_1: FsRow = sheet.Row((_arrow306() + base_row_index) - 1)
                        row_index_set = add(row_1.Index, row_index_set)
                        cell_1: FsCell = row_1.Item(col_i)
                        cell_1.DataType = el_2.fields[0][0]
                        cell_1.Value = el_2.fields[0][1]

                    else: 
                        i_7: RowIndex = el_2.fields[0]
                        row: FsRow = sheet.Row((RowIndex__get_Index(i_7) + base_row_index) - 1)
                        row_index_set = add(RowIndex__get_Index(i_7), row_index_set)
                        cell: FsCell = row.Item(col_i)
                        cell.DataType = el_2.fields[1][0]
                        cell.Value = el_2.fields[1][1]


                iterate(action, elements_2)

            iterate(action_1, columns)

        elif pattern_matching_result == 1:
            max_row: int = (sheet.CellCollection.MaxRowNumber + 1) or 0
            range: FsRangeAddress = FsRangeAddress(FsAddress(max_row, 1), FsAddress(max_row, 1))
            table: FsTable = sheet.Table(name, range)
            FsSpreadsheet_DSL_Workbook__Workbook_parseTable_Static(sheet.CellCollection, table, table_elements)

        elif pattern_matching_result == 2:
            def action_2(_arg_1: SheetElement, _arg: Any=_arg) -> None:
                if _arg_1.tag == 1:
                    row_2: FsRow = sheet.Row(RowIndex__get_Index(_arg_1.fields[0]))
                    FsSpreadsheet_DSL_Workbook__Workbook_parseRow_Static(sheet.CellCollection, row_2, _arg_1.fields[1])

                elif _arg_1.tag == 2:
                    def _arrow307(__unit: None=None, _arg_1: Any=_arg_1) -> int:
                        nonlocal row_index_set
                        i_1: int = 1
                        while FSharpSet__Contains(row_index_set, i_1):
                            i_1 = (i_1 + 1) or 0
                        row_index_set = add(i_1, row_index_set)
                        return i_1

                    row_4: FsRow = sheet.Row(_arrow307())
                    FsSpreadsheet_DSL_Workbook__Workbook_parseRow_Static(sheet.CellCollection, row_4, _arg_1.fields[0])

                else: 
                    raise Exception("Expected row elements")


            iterate(action_2, l_1)

        elif pattern_matching_result == 3:
            to_fail(printf("Invalid sheet element %s"))(s)


    iterate(action_3, split_rows_and_columns(els))


def FsSpreadsheet_DSL_Workbook__Workbook_Parse(self_1: Workbook) -> FsWorkbook:
    workbook: FsWorkbook = FsWorkbook()
    def action(i: int, wb_el: WorkbookElement, self_1: Any=self_1) -> None:
        if wb_el.tag == 1:
            worksheet_1: FsWorksheet = FsWorksheet(wb_el.fields[0])
            FsSpreadsheet_DSL_Workbook__Workbook_parseSheet_Static(worksheet_1, wb_el.fields[1])
            value_1: None = workbook.AddWorksheet(worksheet_1)
            ignore(None)

        else: 
            def _arrow308(__unit: None=None, i: Any=i, wb_el: Any=wb_el) -> str:
                arg: int = (i + 1) or 0
                return to_text(printf("Sheet%i"))(arg)

            worksheet: FsWorksheet = FsWorksheet(_arrow308())
            FsSpreadsheet_DSL_Workbook__Workbook_parseSheet_Static(worksheet, wb_el.fields[0])
            value: None = workbook.AddWorksheet(worksheet)
            ignore(None)


    iterate_indexed(action, self_1.fields[0])
    return workbook


__all__ = ["split_rows_and_columns", "FsSpreadsheet_DSL_Workbook__Workbook_parseTable_Static", "FsSpreadsheet_DSL_Workbook__Workbook_parseRow_Static", "FsSpreadsheet_DSL_Workbook__Workbook_parseSheet_Static", "FsSpreadsheet_DSL_Workbook__Workbook_Parse"]




from __future__ import annotations
from typing import Any
from ...fable_library.option import some
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.util import int32_to_string
from ..fs_address import FsAddress
from .fs_range_address import FsRangeAddress
from .fs_range_base import (FsRangeBase, FsRangeBase_reflection)
from .fs_range_row import FsRangeRow

def _expr217() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRange", None, FsRange, FsRangeBase_reflection())


class FsRange(FsRangeBase):
    def __init__(self, range_address: FsRangeAddress, style_value: Any | None=None) -> None:
        super().__init__(range_address)
        pass

    @staticmethod
    def from_range_base(range_base: FsRangeBase) -> FsRange:
        return FsRange(range_base.RangeAddress, some(None))

    def Row(self, row: int) -> FsRangeRow:
        if True if (row <= 0) else (((row + super().RangeAddress.FirstAddress.RowNumber) - 1) > 1048576):
            raise Exception(int32_to_string(row), to_text(printf("Row number must be between 1 and %i"))(1048576))

        return FsRangeRow(FsRangeAddress(FsAddress((super().RangeAddress.FirstAddress.RowNumber + row) - 1, super().RangeAddress.FirstAddress.ColumnNumber), FsAddress((super().RangeAddress.FirstAddress.RowNumber + row) - 1, super().RangeAddress.LastAddress.ColumnNumber)))

    def FirstRow(self, __unit: None=None) -> FsRangeRow:
        self_1: FsRange = self
        return self_1.Row(1)


FsRange_reflection = _expr217

def FsRange__ctor_Z15E90CDC(range_address: FsRangeAddress, style_value: Any | None=None) -> FsRange:
    return FsRange(range_address, style_value)


__all__ = ["FsRange_reflection"]




from __future__ import annotations
from ..arctrl_core.data_map import DataMap
from ..fable_library.seq import (try_pick, for_all)
from ..fable_library.string_ import (to_fail, printf)
from ..fable_library.types import Array
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .DataMapTable.data_map_table import (try_from_fs_worksheet, to_fs_worksheet)

def from_fs_workbook(doc: FsWorkbook) -> DataMap:
    try: 
        worksheets: Array[FsWorksheet] = doc.GetWorksheets()
        data_map_table: DataMap | None = try_pick(try_from_fs_worksheet, worksheets)
        if data_map_table is None:
            def sheet_is_empty(sheet: FsWorksheet) -> bool:
                return sheet.CellCollection.Count == 0

            if for_all(sheet_is_empty, worksheets):
                return DataMap.init()

            else: 
                raise Exception("No DataMapTable was found in any of the sheets of the workbook")


        else: 
            return data_map_table


    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Could not parse datamap: \n%s"))(arg)



def to_fs_workbook(data_map: DataMap) -> FsWorkbook:
    doc: FsWorkbook = FsWorkbook()
    sheet: FsWorksheet = to_fs_worksheet(data_map)
    doc.AddWorksheet(sheet)
    return doc


__all__ = ["from_fs_workbook", "to_fs_workbook"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.iproperty_value import IPropertyValue
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.value import Value as Value_4
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, IGetters)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .context.rocrate.property_value_context import context_jsonvalue
from .encode import try_include
from .ontology_annotation import AnnotationValue_decoder

__A_ = TypeVar("__A_")

_T = TypeVar("_T")

def gen_id(p: IPropertyValue) -> str:
    match_value: OntologyAnnotation | None = p.GetCategory()
    match_value_1: Value_4 | None = p.GetValue()
    match_value_2: OntologyAnnotation | None = p.GetUnit()
    (pattern_matching_result, t, u, v, t_1, v_1) = (None, None, None, None, None, None)
    if match_value is not None:
        if match_value_1 is not None:
            if match_value_2 is None:
                pattern_matching_result = 1
                t_1 = match_value
                v_1 = match_value_1

            else: 
                pattern_matching_result = 0
                t = match_value
                u = match_value_2
                v = match_value_1


        else: 
            pattern_matching_result = 2


    else: 
        pattern_matching_result = 2

    if pattern_matching_result == 0:
        return ((((((("#" + p.GetAdditionalType()) + "/") + t.NameText) + "=") + v.Text) + "") + u.NameText) + ""

    elif pattern_matching_result == 1:
        return ((((("#" + p.GetAdditionalType()) + "/") + t_1.NameText) + "=") + v_1.Text) + ""

    elif pattern_matching_result == 2:
        return ("#Empty" + p.GetAdditionalType()) + ""



def encoder(pv: IPropertyValue) -> IEncodable:
    pattern_input: tuple[str | None, str | None]
    match_value: OntologyAnnotation | None = pv.GetCategory()
    if match_value is None:
        pattern_input = (None, None)

    else: 
        oa: OntologyAnnotation = match_value
        pattern_input = (oa.Name, oa.TermAccessionNumber)

    pattern_input_1: tuple[IEncodable | None, IEncodable | None]
    match_value_1: Value_4 | None = pv.GetValue()
    if match_value_1 is None:
        pattern_input_1 = (None, None)

    else: 
        v: Value_4 = match_value_1
        if v.tag == 1:
            class ObjectExpr2073(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any], pv: Any=pv) -> Any:
                    return helpers_1.encode_signed_integral_number(v.fields[0])

            pattern_input_1 = (ObjectExpr2073(), None)

        elif v.tag == 2:
            class ObjectExpr2074(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any], pv: Any=pv) -> Any:
                    return helpers_2.encode_decimal_number(v.fields[0])

            pattern_input_1 = (ObjectExpr2074(), None)

        elif v.tag == 0:
            oa_1: OntologyAnnotation = v.fields[0]
            def _arrow2076(value_3: str, pv: Any=pv) -> IEncodable:
                class ObjectExpr2075(IEncodable):
                    def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                        return helpers_3.encode_string(value_3)

                return ObjectExpr2075()

            def _arrow2078(value_5: str, pv: Any=pv) -> IEncodable:
                class ObjectExpr2077(IEncodable):
                    def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                        return helpers_4.encode_string(value_5)

                return ObjectExpr2077()

            pattern_input_1 = (map(_arrow2076, oa_1.Name), map(_arrow2078, oa_1.TermAccessionNumber))

        else: 
            class ObjectExpr2079(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any], pv: Any=pv) -> Any:
                    return helpers.encode_string(v.fields[0])

            pattern_input_1 = (ObjectExpr2079(), None)


    pattern_input_2: tuple[str | None, str | None]
    match_value_2: OntologyAnnotation | None = pv.GetUnit()
    if match_value_2 is None:
        pattern_input_2 = (None, None)

    else: 
        oa_2: OntologyAnnotation = match_value_2
        pattern_input_2 = (oa_2.Name, oa_2.TermAccessionNumber)

    def chooser(tupled_arg: tuple[str, IEncodable | None], pv: Any=pv) -> tuple[str, IEncodable] | None:
        def mapping(v_1_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1_1)

        return map(mapping, tupled_arg[1])

    def _arrow2083(__unit: None=None, pv: Any=pv) -> IEncodable:
        value_8: str = gen_id(pv)
        class ObjectExpr2082(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_8)

        return ObjectExpr2082()

    class ObjectExpr2084(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any], pv: Any=pv) -> Any:
            return helpers_6.encode_string("PropertyValue")

    def _arrow2086(__unit: None=None, pv: Any=pv) -> IEncodable:
        value_10: str = pv.GetAdditionalType()
        class ObjectExpr2085(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                return helpers_7.encode_string(value_10)

        return ObjectExpr2085()

    def _arrow2088(value_11: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2087(IEncodable):
            def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                return helpers_8.encode_string(value_11)

        return ObjectExpr2087()

    def _arrow2090(value_13: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2089(IEncodable):
            def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
                return helpers_9.encode_string(value_13)

        return ObjectExpr2089()

    def _arrow2092(value_15: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2091(IEncodable):
            def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
                return helpers_10.encode_string(value_15)

        return ObjectExpr2091()

    def _arrow2094(value_17: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2093(IEncodable):
            def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
                return helpers_11.encode_string(value_17)

        return ObjectExpr2093()

    def _arrow2096(value_19: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2095(IEncodable):
            def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
                return helpers_12.encode_string(value_19)

        return ObjectExpr2095()

    def _arrow2097(x: IEncodable, pv: Any=pv) -> IEncodable:
        return x

    def _arrow2098(x_1: IEncodable, pv: Any=pv) -> IEncodable:
        return x_1

    def _arrow2100(value_21: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2099(IEncodable):
            def Encode(self, helpers_13: IEncoderHelpers_1[Any]) -> Any:
                return helpers_13.encode_string(value_21)

        return ObjectExpr2099()

    def _arrow2102(value_23: str, pv: Any=pv) -> IEncodable:
        class ObjectExpr2101(IEncodable):
            def Encode(self, helpers_14: IEncoderHelpers_1[Any]) -> Any:
                return helpers_14.encode_string(value_23)

        return ObjectExpr2101()

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2083()), ("@type", ObjectExpr2084()), ("additionalType", _arrow2086()), try_include("alternateName", _arrow2088, pv.AlternateName()), try_include("measurementMethod", _arrow2090, pv.MeasurementMethod()), try_include("description", _arrow2092, pv.Description()), try_include("category", _arrow2094, pattern_input[0]), try_include("categoryCode", _arrow2096, pattern_input[1]), try_include("value", _arrow2097, pattern_input_1[0]), try_include("valueCode", _arrow2098, pattern_input_1[1]), try_include("unit", _arrow2100, pattern_input_2[0]), try_include("unitCode", _arrow2102, pattern_input_2[1]), ("@context", context_jsonvalue)]))
    class ObjectExpr2103(IEncodable):
        def Encode(self, helpers_15: IEncoderHelpers_1[Any], pv: Any=pv) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_15))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_15.encode_object(arg)

    return ObjectExpr2103()


def decoder(create: Callable[[str | None, str | None, str | None, OntologyAnnotation | None, Value_4 | None, OntologyAnnotation | None], _T]) -> Decoder_1[Any]:
    def _arrow2105(get: IGetters, create: Any=create) -> _T:
        alternate_name: str | None
        object_arg: IOptionalGetter = get.Optional
        alternate_name = object_arg.Field("alternateName", string)
        measurement_method: str | None
        object_arg_1: IOptionalGetter = get.Optional
        measurement_method = object_arg_1.Field("measurementMethod", string)
        description: str | None
        object_arg_2: IOptionalGetter = get.Optional
        description = object_arg_2.Field("description", string)
        category: OntologyAnnotation | None
        name: str | None
        object_arg_3: IOptionalGetter = get.Optional
        name = object_arg_3.Field("category", string)
        code: str | None
        object_arg_4: IOptionalGetter = get.Optional
        code = object_arg_4.Field("categoryCode", string)
        (pattern_matching_result, code_1) = (None, None)
        if name is None:
            if code is not None:
                if code == "":
                    pattern_matching_result = 0

                else: 
                    pattern_matching_result = 2
                    code_1 = code


            else: 
                pattern_matching_result = 0


        elif code is not None:
            if code == "":
                pattern_matching_result = 1

            else: 
                pattern_matching_result = 2
                code_1 = code


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            category = None

        elif pattern_matching_result == 1:
            try: 
                category = OntologyAnnotation.create(name)

            except Exception as err:
                raise Exception(((("Error while decoding category (name:" + str(name)) + "): ") + str(err)) + "")


        elif pattern_matching_result == 2:
            try: 
                category = OntologyAnnotation.from_term_annotation(code_1, name)

            except Exception as err_1:
                raise Exception(((((("Error while decoding category (name:" + str(name)) + ", code:") + code_1) + "): ") + str(err_1)) + "")


        unit: OntologyAnnotation | None
        name_1: str | None
        object_arg_5: IOptionalGetter = get.Optional
        name_1 = object_arg_5.Field("unit", string)
        code_2: str | None
        object_arg_6: IOptionalGetter = get.Optional
        code_2 = object_arg_6.Field("unitCode", string)
        (pattern_matching_result_1, code_3) = (None, None)
        if name_1 is None:
            if code_2 is not None:
                if code_2 == "":
                    pattern_matching_result_1 = 0

                else: 
                    pattern_matching_result_1 = 2
                    code_3 = code_2


            else: 
                pattern_matching_result_1 = 0


        elif code_2 is not None:
            if code_2 == "":
                pattern_matching_result_1 = 1

            else: 
                pattern_matching_result_1 = 2
                code_3 = code_2


        else: 
            pattern_matching_result_1 = 1

        if pattern_matching_result_1 == 0:
            unit = None

        elif pattern_matching_result_1 == 1:
            try: 
                unit = OntologyAnnotation.create(name_1)

            except Exception as err_2:
                raise Exception(((("Error while decoding unit (name:" + str(name_1)) + "): ") + str(err_2)) + "")


        elif pattern_matching_result_1 == 2:
            try: 
                unit = OntologyAnnotation.from_term_annotation(code_3, name_1)

            except Exception as err_3:
                raise Exception(((((("Error while decoding unit (name:" + str(name_1)) + ", code:") + code_3) + "): ") + str(err_3)) + "")


        def _arrow2104(__unit: None=None) -> Value_4 | None:
            value: str | None
            object_arg_7: IOptionalGetter = get.Optional
            value = object_arg_7.Field("value", AnnotationValue_decoder)
            code_4: str | None
            object_arg_8: IOptionalGetter = get.Optional
            code_4 = object_arg_8.Field("valueCode", string)
            if (code_4 is None) if (value is None) else False:
                return None

            else: 
                try: 
                    return Value_4.from_options(value, None, code_4)

                except Exception as err_4:
                    raise Exception(((((("Error while decoding value " + str(value)) + ",") + str(code_4)) + ": ") + str(err_4)) + "")



        return create(alternate_name, measurement_method, description, category, _arrow2104(), unit)

    return object(_arrow2105)


__all__ = ["gen_id", "encoder", "decoder"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.value import Value
from ...fable_library.list import of_array
from ...thoth_json_core.decode import (one_of, map, int_1, float_1, string)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..ontology_annotation import (OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)

def encoder(id_map: Any | None, value: Value) -> IEncodable:
    if value.tag == 1:
        class ObjectExpr2229(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], id_map: Any=id_map, value: Any=value) -> Any:
                return helpers_1.encode_signed_integral_number(value.fields[0])

        return ObjectExpr2229()

    elif value.tag == 3:
        class ObjectExpr2230(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], id_map: Any=id_map, value: Any=value) -> Any:
                return helpers_2.encode_string(value.fields[0])

        return ObjectExpr2230()

    elif value.tag == 0:
        return OntologyAnnotation_ISAJson_encoder(id_map, value.fields[0])

    else: 
        class ObjectExpr2231(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], id_map: Any=id_map, value: Any=value) -> Any:
                return helpers.encode_decimal_number(value.fields[0])

        return ObjectExpr2231()



def _arrow2233(Item: int) -> Value:
    return Value(1, Item)


def _arrow2234(Item_1: float) -> Value:
    return Value(2, Item_1)


def _arrow2235(Item_2: OntologyAnnotation) -> Value:
    return Value(0, Item_2)


def _arrow2236(Item_3: str) -> Value:
    return Value(3, Item_3)


decoder: Decoder_1[Value] = one_of(of_array([map(_arrow2233, int_1), map(_arrow2234, float_1), map(_arrow2235, OntologyAnnotation_ISAJson_decoder), map(_arrow2236, string)]))

__all__ = ["encoder", "decoder"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.arc_types import ArcInvestigation
from ..arctrl_file_system.path import combine_many
from ..arctrl_spreadsheet.arc_investigation import (ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF, ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D)
from ..fable_library.array_ import equals_with
from ..fable_library.types import Array
from .contract import (Contract, DTOType, DTO)

def _007CInvestigationPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3753(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 1) if (not equals_with(_arrow3753, input, None)) else False:
        if input[0] == "isa.investigation.xlsx":
            pattern_matching_result = 0

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return combine_many(input)

    elif pattern_matching_result == 1:
        return None



def ARCtrl_ArcInvestigation__ArcInvestigation_ToCreateContract(this: ArcInvestigation) -> Contract:
    return Contract.create_create("isa.investigation.xlsx", DTOType(4), DTO(0, ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(this)))


def ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract(this: ArcInvestigation) -> Contract:
    return Contract.create_update("isa.investigation.xlsx", DTOType(4), DTO(0, ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(this)))


def ARCtrl_ArcInvestigation__ArcInvestigation_toCreateContract_Static_Z720BD3FF(inv: ArcInvestigation) -> Contract:
    return ARCtrl_ArcInvestigation__ArcInvestigation_ToCreateContract(inv)


def ARCtrl_ArcInvestigation__ArcInvestigation_toUpdateContract_Static_Z720BD3FF(inv: ArcInvestigation) -> Contract:
    return ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract(inv)


def ARCtrl_ArcInvestigation__ArcInvestigation_tryFromReadContract_Static_7570923F(c: Contract) -> ArcInvestigation | None:
    (pattern_matching_result, fsworkbook) = (None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 4:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        pattern_matching_result = 0
                        fsworkbook = c.DTO.fields[0]

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D(fsworkbook)

    elif pattern_matching_result == 1:
        return None



__all__ = ["_007CInvestigationPath_007C__007C", "ARCtrl_ArcInvestigation__ArcInvestigation_ToCreateContract", "ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract", "ARCtrl_ArcInvestigation__ArcInvestigation_toCreateContract_Static_Z720BD3FF", "ARCtrl_ArcInvestigation__ArcInvestigation_toUpdateContract_Static_Z720BD3FF", "ARCtrl_ArcInvestigation__ArcInvestigation_tryFromReadContract_Static_7570923F"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.list import (FSharpList, try_find, exists, append, singleton, filter, empty)
from ...fable_library.option import (map, default_arg)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, list_type, record_type)
from ...fable_library.types import Record
from ...fable_library.util import equals
from ..comment import (Comment, Comment_reflection)
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from .component import (Component, Component_reflection)
from .protocol_parameter import (ProtocolParameter, ProtocolParameter_reflection)

def _expr763() -> TypeInfo:
    return record_type("ARCtrl.Process.Protocol", [], Protocol, lambda: [("ID", option_type(string_type)), ("Name", option_type(string_type)), ("ProtocolType", option_type(OntologyAnnotation_reflection())), ("Description", option_type(string_type)), ("Uri", option_type(string_type)), ("Version", option_type(string_type)), ("Parameters", option_type(list_type(ProtocolParameter_reflection()))), ("Components", option_type(list_type(Component_reflection()))), ("Comments", option_type(list_type(Comment_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Protocol(Record):
    ID: str | None
    Name: str | None
    ProtocolType: OntologyAnnotation | None
    Description: str | None
    Uri: str | None
    Version: str | None
    Parameters: FSharpList[ProtocolParameter] | None
    Components: FSharpList[Component] | None
    Comments: FSharpList[Comment] | None

Protocol_reflection = _expr763

def Protocol_make(id: str | None=None, name: str | None=None, protocol_type: OntologyAnnotation | None=None, description: str | None=None, uri: str | None=None, version: str | None=None, parameters: FSharpList[ProtocolParameter] | None=None, components: FSharpList[Component] | None=None, comments: FSharpList[Comment] | None=None) -> Protocol:
    return Protocol(id, name, protocol_type, description, uri, version, parameters, components, comments)


def Protocol_create_Z414665E7(Id: str | None=None, Name: str | None=None, ProtocolType: OntologyAnnotation | None=None, Description: str | None=None, Uri: str | None=None, Version: str | None=None, Parameters: FSharpList[ProtocolParameter] | None=None, Components: FSharpList[Component] | None=None, Comments: FSharpList[Comment] | None=None) -> Protocol:
    return Protocol_make(Id, Name, ProtocolType, Description, Uri, Version, Parameters, Components, Comments)


def Protocol_get_empty(__unit: None=None) -> Protocol:
    return Protocol_create_Z414665E7()


def Protocol_tryGetByName(name: str, protocols: FSharpList[Protocol]) -> Protocol | None:
    def _arrow764(p: Protocol, name: Any=name, protocols: Any=protocols) -> bool:
        return equals(p.Name, name)

    return try_find(_arrow764, protocols)


def Protocol_existsByName(name: str, protocols: FSharpList[Protocol]) -> bool:
    def _arrow765(p: Protocol, name: Any=name, protocols: Any=protocols) -> bool:
        return equals(p.Name, name)

    return exists(_arrow765, protocols)


def Protocol_add(protocols: FSharpList[Protocol], protocol: Protocol) -> FSharpList[Protocol]:
    return append(protocols, singleton(protocol))


def Protocol_removeByName(name: str, protocols: FSharpList[Protocol]) -> FSharpList[Protocol]:
    def _arrow766(p: Protocol, name: Any=name, protocols: Any=protocols) -> bool:
        return not equals(p.Name, name)

    return filter(_arrow766, protocols)


def Protocol_getComments_3BF20962(protocol: Protocol) -> FSharpList[Comment] | None:
    return protocol.Comments


def Protocol_mapComments(f: Callable[[FSharpList[Comment]], FSharpList[Comment]], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, map(f, protocol.Comments))


def Protocol_setComments(protocol: Protocol, comments: FSharpList[Comment]) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, comments)


def Protocol_getProtocolType_3BF20962(protocol: Protocol) -> OntologyAnnotation | None:
    return protocol.ProtocolType


def Protocol_mapProtocolType(f: Callable[[OntologyAnnotation], OntologyAnnotation], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, map(f, protocol.ProtocolType), protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_setProtocolType(protocol: Protocol, protocol_type: OntologyAnnotation) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol_type, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_getVersion_3BF20962(protocol: Protocol) -> str | None:
    return protocol.Version


def Protocol_mapVersion(f: Callable[[str], str], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, map(f, protocol.Version), protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_setVersion(protocol: Protocol, version: str) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_getName_3BF20962(protocol: Protocol) -> str | None:
    return protocol.Name


def Protocol_mapName(f: Callable[[str], str], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, map(f, protocol.Name), protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_setName(protocol: Protocol, name: str) -> Protocol:
    return Protocol(protocol.ID, name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_getDescription_3BF20962(protocol: Protocol) -> str | None:
    return protocol.Description


def Protocol_mapDescription(f: Callable[[str], str], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, map(f, protocol.Description), protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_setDescription(protocol: Protocol, description: str) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, description, protocol.Uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_getUri_3BF20962(protocol: Protocol) -> str | None:
    return protocol.Uri


def Protocol_mapUri(f: Callable[[str], str], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, map(f, protocol.Uri), protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_setUri(protocol: Protocol, uri: str) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, uri, protocol.Version, protocol.Parameters, protocol.Components, protocol.Comments)


def Protocol_getComponents_3BF20962(protocol: Protocol) -> FSharpList[Component] | None:
    return protocol.Components


def Protocol_mapComponents(f: Callable[[FSharpList[Component]], FSharpList[Component]], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, map(f, protocol.Components), protocol.Comments)


def Protocol_setComponents(protocol: Protocol, components: FSharpList[Component]) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, protocol.Parameters, components, protocol.Comments)


def Protocol_addComponent(comp: Component, protocol: Protocol) -> Protocol:
    return Protocol_setComponents(protocol, append(default_arg(protocol.Components, empty()), singleton(comp)))


def Protocol_getParameters_3BF20962(protocol: Protocol) -> FSharpList[ProtocolParameter] | None:
    return protocol.Parameters


def Protocol_mapParameters(f: Callable[[FSharpList[ProtocolParameter]], FSharpList[ProtocolParameter]], protocol: Protocol) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, map(f, protocol.Parameters), protocol.Components, protocol.Comments)


def Protocol_setParameters(protocol: Protocol, parameters: FSharpList[ProtocolParameter]) -> Protocol:
    return Protocol(protocol.ID, protocol.Name, protocol.ProtocolType, protocol.Description, protocol.Uri, protocol.Version, parameters, protocol.Components, protocol.Comments)


def Protocol_addParameter(parameter: ProtocolParameter, protocol: Protocol) -> Protocol:
    return Protocol_setParameters(protocol, append(default_arg(protocol.Parameters, empty()), singleton(parameter)))


__all__ = ["Protocol_reflection", "Protocol_make", "Protocol_create_Z414665E7", "Protocol_get_empty", "Protocol_tryGetByName", "Protocol_existsByName", "Protocol_add", "Protocol_removeByName", "Protocol_getComments_3BF20962", "Protocol_mapComments", "Protocol_setComments", "Protocol_getProtocolType_3BF20962", "Protocol_mapProtocolType", "Protocol_setProtocolType", "Protocol_getVersion_3BF20962", "Protocol_mapVersion", "Protocol_setVersion", "Protocol_getName_3BF20962", "Protocol_mapName", "Protocol_setName", "Protocol_getDescription_3BF20962", "Protocol_mapDescription", "Protocol_setDescription", "Protocol_getUri_3BF20962", "Protocol_mapUri", "Protocol_setUri", "Protocol_getComponents_3BF20962", "Protocol_mapComponents", "Protocol_setComponents", "Protocol_addComponent", "Protocol_getParameters_3BF20962", "Protocol_mapParameters", "Protocol_setParameters", "Protocol_addParameter"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import (value as value_1, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1559() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.Sample", None, Sample, LDNode_reflection())


class Sample(LDNode):
    def __init__(self, id: str, name: Any=None, additional_type: Array[str] | None=None, additional_property: Any | None=None, derives_from: Any | None=None) -> None:
        super().__init__(id, ["bioschemas.org/Sample"], default_arg(additional_type, []))
        this: FSharpRef[Sample] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_property("name", name, this.contents)
        set_optional_property("additionalProperty", additional_property, this.contents)
        set_optional_property("derivesFrom", derives_from, this.contents)

    def GetName(self, __unit: None=None) -> str:
        this: Sample = self
        obj: DynamicObj = this
        if try_get_property_value("name", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("name")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "name") + "\' is set on this \'") + "Sample") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "name") + "\' set on this \'") + "Sample") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_name() -> Callable[[Sample], str]:
        def _arrow1558(s: Sample) -> str:
            return s.GetName()

        return _arrow1558


Sample_reflection = _expr1559

def Sample__ctor_Z502AA21F(id: str, name: Any=None, additional_type: Array[str] | None=None, additional_property: Any | None=None, derives_from: Any | None=None) -> Sample:
    return Sample(id, name, additional_type, additional_property, derives_from)


__all__ = ["Sample_reflection"]




from __future__ import annotations
from ...fable_library.reflection import (TypeInfo, class_type)
from ..Ranges.fs_range_address import FsRangeAddress
from ..Ranges.fs_range_row import (FsRangeRow, FsRangeRow_reflection)

def _expr234() -> TypeInfo:
    return class_type("FsSpreadsheet.FsTableRow", None, FsTableRow, FsRangeRow_reflection())


class FsTableRow(FsRangeRow):
    def __init__(self, range_address: FsRangeAddress) -> None:
        super().__init__(range_address)
        pass


FsTableRow_reflection = _expr234

def FsTableRow__ctor_6A2513BC(range_address: FsRangeAddress) -> FsTableRow:
    return FsTableRow(range_address)


__all__ = ["FsTableRow_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.string_ import (to_text, printf, trim, trim_start as trim_start_1, starts_with_exact, substring)
from ..fable_library.types import (Array, to_string)
from .comment import Comment
from .data_file import (DataFile, DataFile__get_AsString)
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq, hash_1)

def DataAux_nameFromPathAndSelector(path: str, selector: str) -> str:
    return to_text(printf("%s#%s"))(path)(selector)


def DataAux_pathAndSelectorFromName(name: str) -> tuple[str, str | None]:
    name_1: str = trim(name, "#")
    parts: Array[str] = name_1.split("#")
    if len(parts) == 2:
        return (parts[0], parts[1])

    else: 
        return (name_1, None)



def _expr728() -> TypeInfo:
    return class_type("ARCtrl.Data", None, Data)


class Data:
    def __init__(self, id: str | None=None, name: str | None=None, data_type: DataFile | None=None, format: str | None=None, selector_format: str | None=None, comments: Array[Comment] | None=None) -> None:
        self._id: str | None = id
        pattern_input_1: tuple[str | None, str | None]
        if name is None:
            pattern_input_1 = (None, None)

        else: 
            pattern_input: tuple[str, str | None] = DataAux_pathAndSelectorFromName(name)
            pattern_input_1 = (pattern_input[0], pattern_input[1])

        self._selector: str | None = pattern_input_1[1]
        self._filePath: str | None = pattern_input_1[0]
        self._dataType: DataFile | None = data_type
        self._format: str | None = format
        self._selectorFormat: str | None = selector_format
        self._comments: Array[Comment] = default_arg(comments, [])

    @property
    def ID(self, __unit: None=None) -> str | None:
        this: Data = self
        return this._id

    @ID.setter
    def ID(self, id: str | None=None) -> None:
        this: Data = self
        this._id = id

    @property
    def Name(self, __unit: None=None) -> str | None:
        this: Data = self
        match_value: str | None = this._filePath
        match_value_1: str | None = this._selector
        def _arrow723(__unit: None=None) -> str | None:
            p_1: str = match_value
            return p_1

        def _arrow724(__unit: None=None) -> str | None:
            p: str = match_value
            s: str = match_value_1
            return DataAux_nameFromPathAndSelector(p, s)

        return None if (match_value is None) else (_arrow723() if (match_value_1 is None) else _arrow724())

    @Name.setter
    def Name(self, name: str | None=None) -> None:
        this: Data = self
        if name is None:
            this._filePath = None
            this._selector = None

        else: 
            pattern_input: tuple[str, str | None] = DataAux_pathAndSelectorFromName(name)
            this._filePath = pattern_input[0]
            this._selector = pattern_input[1]


    @property
    def FilePath(self, __unit: None=None) -> str | None:
        this: Data = self
        return this._filePath

    @FilePath.setter
    def FilePath(self, file_path: str | None=None) -> None:
        this: Data = self
        this._filePath = file_path

    @property
    def Selector(self, __unit: None=None) -> str | None:
        this: Data = self
        return this._selector

    @Selector.setter
    def Selector(self, selector: str | None=None) -> None:
        this: Data = self
        this._selector = selector

    @property
    def DataType(self, __unit: None=None) -> DataFile | None:
        this: Data = self
        return this._dataType

    @DataType.setter
    def DataType(self, data_type: DataFile | None=None) -> None:
        this: Data = self
        this._dataType = data_type

    @property
    def Format(self, __unit: None=None) -> str | None:
        this: Data = self
        return this._format

    @Format.setter
    def Format(self, format: str | None=None) -> None:
        this: Data = self
        this._format = format

    @property
    def SelectorFormat(self, __unit: None=None) -> str | None:
        this: Data = self
        return this._selectorFormat

    @SelectorFormat.setter
    def SelectorFormat(self, selector_format: str | None=None) -> None:
        this: Data = self
        this._selectorFormat = selector_format

    @property
    def Comments(self, __unit: None=None) -> Array[Comment]:
        this: Data = self
        return this._comments

    @Comments.setter
    def Comments(self, comments: Array[Comment]) -> None:
        this: Data = self
        this._comments = comments

    @staticmethod
    def make(id: str | None=None, name: str | None=None, data_type: DataFile | None=None, format: str | None=None, selector_format: str | None=None, comments: Array[Comment] | None=None) -> Data:
        return Data(id, name, data_type, format, selector_format, comments)

    @staticmethod
    def create(Id: str | None=None, Name: str | None=None, DataType: DataFile | None=None, Format: str | None=None, SelectorFormat: str | None=None, Comments: Array[Comment] | None=None) -> Data:
        return Data.make(Id, Name, DataType, Format, SelectorFormat, Comments)

    @staticmethod
    def empty() -> Data:
        return Data.create()

    @property
    def NameText(self, __unit: None=None) -> str:
        this: Data = self
        return default_arg(this.Name, "")

    def GetAbsolutePathBy(self, f: Callable[[str], str], check_existence_from_root: Callable[[str], bool] | None=None) -> str:
        this: Data = self
        def _arrow725(_arg: str) -> bool:
            return False

        check_existence_from_root_1: Callable[[str], bool] = default_arg(check_existence_from_root, _arrow725)
        match_value: str | None = this.FilePath
        if match_value is None:
            raise Exception("Data does not have a file path")

        else: 
            p_3: str
            s: str = trim_start_1(match_value, "/")
            p_3 = substring(s, len("./")) if starts_with_exact(s, "./") else s
            def _arrow726(__unit: None=None) -> bool:
                p: str = p_3
                return True if (True if (True if starts_with_exact(p, "assays/") else starts_with_exact(p, "studies/")) else starts_with_exact(p, "workflows/")) else starts_with_exact(p, "runs/")

            def _arrow727(__unit: None=None) -> bool:
                p_1: str = p_3
                return True if starts_with_exact(p_1, "http:") else starts_with_exact(p_1, "https:")

            return p_3 if (True if (True if check_existence_from_root_1(p_3) else _arrow726()) else _arrow727()) else f(p_3)


    def GetAbsolutePathForAssay(self, assay_identifier: str, check_existence_from_root: Callable[[str], bool] | None=None) -> str:
        this: Data = self
        def f(p: str) -> str:
            return (("assays/" + assay_identifier) + "/dataset/") + trim_start_1(p, "/")

        return this.GetAbsolutePathBy(f, check_existence_from_root)

    def GetAbsolutePathForStudy(self, study_identifier: str, check_existence_from_root: Callable[[str], bool] | None=None) -> str:
        this: Data = self
        def f(p: str) -> str:
            return (("studies/" + study_identifier) + "/resources/") + trim_start_1(p, "/")

        return this.GetAbsolutePathBy(f, check_existence_from_root)

    def Copy(self, __unit: None=None) -> Data:
        this: Data = self
        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        return Data(this.ID, this.Name, this.DataType, this.Format, this.SelectorFormat, next_comments)

    def __hash__(self, __unit: None=None) -> Any:
        this: Data = self
        return box_hash_array([box_hash_option(this.ID), box_hash_option(this.Name), box_hash_option(this.DataType), box_hash_option(this.Format), box_hash_option(this.SelectorFormat), box_hash_seq(this.Comments)])

    def __eq__(self, obj: Any=None) -> bool:
        this: Data = self
        return hash_1(this) == hash_1(obj)

    def Print(self, __unit: None=None) -> str:
        this: Data = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Data = self
        match_value: DataFile | None = this.DataType
        if match_value is None:
            arg_2: str = this.NameText
            return to_text(printf("%s"))(arg_2)

        else: 
            t: DataFile = match_value
            arg: str = this.NameText
            arg_1: str = DataFile__get_AsString(t)
            return to_text(printf("%s [%s]"))(arg)(arg_1)



Data_reflection = _expr728

def Data__ctor_5909441C(id: str | None=None, name: str | None=None, data_type: DataFile | None=None, format: str | None=None, selector_format: str | None=None, comments: Array[Comment] | None=None) -> Data:
    return Data(id, name, data_type, format, selector_format, comments)


__all__ = ["DataAux_nameFromPathAndSelector", "DataAux_pathAndSelectorFromName", "Data_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.list import (FSharpList, map as map_1, singleton, of_array)
from ..fable_library.option import (default_arg, map)
from ..fable_library.seq import (for_all, to_list, delay, map as map_2, collect, singleton as singleton_1)
from ..fable_library.string_ import join
from ..fable_library.util import IEnumerable_1
from .preprocessing import write as write_1
from .yamlicious_types import (YAMLElement, YAMLContent, PreprocessorElement, Config)

def StyleVerifier_checkInlineSequence(ele: FSharpList[YAMLElement]) -> bool:
    def predicate(x: YAMLElement, ele: Any=ele) -> bool:
        (pattern_matching_result,) = (None,)
        if x.tag == 1:
            if x.fields[0].Comment is None:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False


    return for_all(predicate, ele)


def Formatting_mkComment(comment: str) -> str:
    return "#" + comment


def Formatting_mkKey(key: str) -> str:
    return key + ":"


def Formatting_mkContent(content: YAMLContent) -> str:
    def mapping(s: str, content: Any=content) -> str:
        return " " + Formatting_mkComment(s)

    return content.Value + default_arg(map(mapping, content.Comment), "")


def Formatting_mkInlineSequence(seq: FSharpList[YAMLElement]) -> str:
    def mapping(x: YAMLElement, seq: Any=seq) -> str:
        (pattern_matching_result, v) = (None, None)
        if x.tag == 1:
            if x.fields[0].Comment is None:
                pattern_matching_result = 0
                v = x.fields[0].Value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return v

        elif pattern_matching_result == 1:
            raise Exception("Invalid sequence element")


    return ("[" + join(", ", map_1(mapping, seq))) + "]"


def Formatting_mkMinusLine(c: str) -> str:
    return "- " + c


def detokenize(ele: YAMLElement) -> PreprocessorElement:
    def loop(ele_1: YAMLElement, ele: Any=ele) -> PreprocessorElement:
        if ele_1.tag == 0:
            v: YAMLElement = ele_1.fields[1]
            key: YAMLContent = ele_1.fields[0]
            (pattern_matching_result, v_1, seq_1, any_else) = (None, None, None, None)
            if v.tag == 1:
                pattern_matching_result = 0
                v_1 = v.fields[0]

            elif v.tag == 2:
                if StyleVerifier_checkInlineSequence(v.fields[0]):
                    pattern_matching_result = 1
                    seq_1 = v.fields[0]

                else: 
                    pattern_matching_result = 2
                    any_else = v


            else: 
                pattern_matching_result = 2
                any_else = v

            if pattern_matching_result == 0:
                return PreprocessorElement(2, (Formatting_mkKey(key.Value) + " ") + Formatting_mkContent(v_1))

            elif pattern_matching_result == 1:
                return PreprocessorElement(2, (Formatting_mkKey(key.Value) + " ") + Formatting_mkInlineSequence(seq_1))

            elif pattern_matching_result == 2:
                return PreprocessorElement(0, of_array([PreprocessorElement(2, Formatting_mkKey(Formatting_mkContent(key))), PreprocessorElement(1, singleton(loop(any_else)))]))


        elif ele_1.tag == 1:
            return PreprocessorElement(2, Formatting_mkContent(ele_1.fields[0]))

        elif ele_1.tag == 3:
            def _arrow408(__unit: None=None, ele_1: Any=ele_1) -> IEnumerable_1[PreprocessorElement]:
                return map_2(loop, ele_1.fields[0])

            return PreprocessorElement(0, to_list(delay(_arrow408)))

        elif ele_1.tag == 2:
            def _arrow410(__unit: None=None, ele_1: Any=ele_1) -> IEnumerable_1[PreprocessorElement]:
                def _arrow409(ele_3: YAMLElement) -> IEnumerable_1[PreprocessorElement]:
                    match_value: YAMLElement = ele_3
                    (pattern_matching_result_1, v_3, seq_5, any_else_1) = (None, None, None, None)
                    if match_value.tag == 1:
                        pattern_matching_result_1 = 0
                        v_3 = match_value.fields[0]

                    elif match_value.tag == 2:
                        if StyleVerifier_checkInlineSequence(match_value.fields[0]):
                            pattern_matching_result_1 = 1
                            seq_5 = match_value.fields[0]

                        else: 
                            pattern_matching_result_1 = 2
                            any_else_1 = match_value


                    else: 
                        pattern_matching_result_1 = 2
                        any_else_1 = match_value

                    if pattern_matching_result_1 == 0:
                        return singleton_1(PreprocessorElement(2, Formatting_mkMinusLine(Formatting_mkContent(v_3))))

                    elif pattern_matching_result_1 == 1:
                        return singleton_1(PreprocessorElement(2, Formatting_mkMinusLine(Formatting_mkInlineSequence(seq_5))))

                    elif pattern_matching_result_1 == 2:
                        return singleton_1(PreprocessorElement(0, of_array([PreprocessorElement(2, "-"), PreprocessorElement(1, singleton(loop(any_else_1)))])))


                return collect(_arrow409, ele_1.fields[0])

            return PreprocessorElement(0, to_list(delay(_arrow410)))

        elif ele_1.tag == 4:
            return PreprocessorElement(2, Formatting_mkComment(ele_1.fields[0]))

        else: 
            return PreprocessorElement(3)


    return loop(ele)


def write(ele: YAMLElement, fconfig: Callable[[Config], Config] | None=None) -> str:
    return write_1(detokenize(ele), fconfig)


__all__ = ["StyleVerifier_checkInlineSequence", "Formatting_mkComment", "Formatting_mkKey", "Formatting_mkContent", "Formatting_mkInlineSequence", "Formatting_mkMinusLine", "detokenize", "write"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.option import (map, bind, default_arg)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ...fable_library.util import int32_to_string
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from ..value import (Value as Value_1, Value_reflection)
from .factor import (Factor, Factor_reflection)

def _expr749() -> TypeInfo:
    return record_type("ARCtrl.Process.FactorValue", [], FactorValue, lambda: [("ID", option_type(string_type)), ("Category", option_type(Factor_reflection())), ("Value", option_type(Value_reflection())), ("Unit", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class FactorValue(Record):
    ID: str | None
    Category: Factor | None
    Value: Value_1 | None
    Unit: OntologyAnnotation | None
    def Print(self, __unit: None=None) -> str:
        this: FactorValue = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: FactorValue = self
        def mapping(f: Factor) -> str:
            return f.NameText

        category: str | None = map(mapping, this.Category)
        def mapping_1(oa: OntologyAnnotation) -> str:
            return oa.NameText

        unit: str | None = map(mapping_1, this.Unit)
        def mapping_2(v: Value_1) -> str:
            s: str = v.PrintCompact()
            if unit is None:
                return s

            else: 
                return (s + " ") + unit


        value: str | None = map(mapping_2, this.Value)
        def _arrow746(__unit: None=None) -> str:
            value_2: str = value
            return value_2

        def _arrow747(__unit: None=None) -> str:
            category_2: str = category
            return (category_2 + ":") + "No Value"

        def _arrow748(__unit: None=None) -> str:
            category_1: str = category
            value_1: str = value
            return (category_1 + ":") + value_1

        return ("" if (value is None) else _arrow746()) if (category is None) else (_arrow747() if (value is None) else _arrow748())

    def AlternateName(self, __unit: None=None) -> str | None:
        return None

    def MeasurementMethod(self, __unit: None=None) -> str | None:
        return None

    def Description(self, __unit: None=None) -> str | None:
        return None

    def GetCategory(self, __unit: None=None) -> OntologyAnnotation | None:
        this: FactorValue = self
        def binder(f: Factor) -> OntologyAnnotation | None:
            return f.FactorType

        return bind(binder, this.Category)

    def GetValue(self, __unit: None=None) -> Value_1 | None:
        this: FactorValue = self
        return this.Value

    def GetUnit(self, __unit: None=None) -> OntologyAnnotation | None:
        this: FactorValue = self
        return this.Unit

    def GetAdditionalType(self, __unit: None=None) -> str:
        return "FactorValue"


FactorValue_reflection = _expr749

def FactorValue_make(id: str | None=None, category: Factor | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> FactorValue:
    return FactorValue(id, category, value, unit)


def FactorValue_create_30BDC49(Id: str | None=None, Category: Factor | None=None, Value: Value_1 | None=None, Unit: OntologyAnnotation | None=None) -> FactorValue:
    return FactorValue_make(Id, Category, Value, Unit)


def FactorValue_get_empty(__unit: None=None) -> FactorValue:
    return FactorValue_create_30BDC49()


def FactorValue__get_ValueText(this: FactorValue) -> str:
    def mapping(oa: Value_1, this: Any=this) -> str:
        if oa.tag == 2:
            return to_string(oa.fields[0])

        elif oa.tag == 1:
            return int32_to_string(oa.fields[0])

        elif oa.tag == 3:
            return oa.fields[0]

        else: 
            return oa.fields[0].NameText


    return default_arg(map(mapping, this.Value), "")


def FactorValue__get_ValueWithUnitText(this: FactorValue) -> str:
    def mapping(oa: OntologyAnnotation, this: Any=this) -> str:
        return oa.NameText

    unit: str | None = map(mapping, this.Unit)
    v: str = FactorValue__get_ValueText(this)
    if unit is None:
        return v

    else: 
        u: str = unit
        return to_text(printf("%s %s"))(v)(u)



def FactorValue__get_NameText(this: FactorValue) -> str:
    def mapping(factor: Factor, this: Any=this) -> str:
        return factor.NameText

    return default_arg(map(mapping, this.Category), "")


def FactorValue__MapCategory_658CFBF6(this: FactorValue, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> FactorValue:
    def mapping(p: Factor, this: Any=this, f: Any=f) -> Factor:
        return p.MapCategory(f)

    return FactorValue(this.ID, map(mapping, this.Category), this.Value, this.Unit)


def FactorValue__SetCategory_ZDED3A0F(this: FactorValue, c: OntologyAnnotation) -> FactorValue:
    def _arrow750(__unit: None=None, this: Any=this, c: Any=c) -> Factor | None:
        match_value: Factor | None = this.Category
        if match_value is None:
            return Factor.create(None, c)

        else: 
            p: Factor = match_value
            return p.SetCategory(c)


    return FactorValue(this.ID, _arrow750(), this.Value, this.Unit)


def FactorValue_getNameAsString_7105C732(fv: FactorValue) -> str:
    return FactorValue__get_NameText(fv)


def FactorValue_nameEqualsString(name: str, fv: FactorValue) -> bool:
    return FactorValue__get_NameText(fv) == name


def FactorValue_createAsPV(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> FactorValue:
    def mapping(c: OntologyAnnotation, alternate_name: Any=alternate_name, measurement_method: Any=measurement_method, description: Any=description, category: Any=category, value: Any=value, unit: Any=unit) -> Factor:
        return Factor.create(None, c)

    return FactorValue_create_30BDC49(None, map(mapping, category), value, unit)


__all__ = ["FactorValue_reflection", "FactorValue_make", "FactorValue_create_30BDC49", "FactorValue_get_empty", "FactorValue__get_ValueText", "FactorValue__get_ValueWithUnitText", "FactorValue__get_NameText", "FactorValue__MapCategory_658CFBF6", "FactorValue__SetCategory_ZDED3A0F", "FactorValue_getNameAsString_7105C732", "FactorValue_nameEqualsString", "FactorValue_createAsPV"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_core.arc_types import ArcWorkflow
from ..arctrl_core.comment import Remark
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.person import Person
from ..fable_library.list import (FSharpList, empty, of_seq)
from ..fable_library.seq import (delay, append, singleton, iterate_indexed, map, try_find, try_pick)
from ..fable_library.string_ import (to_fail, printf, to_console)
from ..fable_library.util import (get_enumerator, IEnumerable_1, IEnumerator, to_enumerable)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .DataMapTable.data_map_table import try_from_fs_worksheet
from .Metadata.contacts import (from_rows as from_rows_1, to_rows as to_rows_1)
from .Metadata.sparse_table import (SparseRowModule_tryGetValueAt, SparseRowModule_fromValues, SparseRowModule_writeToSheet, SparseRowModule_fromFsRow, SparseRowModule_getAllValues, SparseRowModule_fromAllValues)
from .Metadata.workflow import (from_rows, to_rows)

def ArcWorkflow_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> ArcWorkflow:
    en: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(rows)
    def loop(last_row_mut: str | None, workflow_mut: ArcWorkflow | None, contacts_mut: FSharpList[Person], row_number_mut: int, rows: Any=rows) -> ArcWorkflow:
        while True:
            (last_row, workflow, contacts, row_number) = (last_row_mut, workflow_mut, contacts_mut, row_number_mut)
            (pattern_matching_result,) = (None,)
            if last_row is not None:
                if last_row == "WORKFLOW":
                    pattern_matching_result = 0

                elif last_row == "WORKFLOW CONTACTS":
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                pattern_input: tuple[str | None, int, FSharpList[Remark], ArcWorkflow] = from_rows(row_number + 1, en)
                last_row_mut = pattern_input[0]
                workflow_mut = pattern_input[3]
                contacts_mut = contacts
                row_number_mut = pattern_input[1]
                continue

            elif pattern_matching_result == 1:
                pattern_input_1: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_1("Workflow Person", row_number + 1, en)
                last_row_mut = pattern_input_1[0]
                workflow_mut = workflow
                contacts_mut = pattern_input_1[3]
                row_number_mut = pattern_input_1[1]
                continue

            elif pattern_matching_result == 2:
                if workflow is not None:
                    workflow_2: ArcWorkflow = workflow
                    workflow_2.Contacts = list(contacts)
                    return workflow_2

                else: 
                    return ArcWorkflow.create(create_missing_identifier(), None, None, None, None, None, None, None, None, None, list(contacts))


            break

    if en.System_Collections_IEnumerator_MoveNext():
        return loop(SparseRowModule_tryGetValueAt(0, en.System_Collections_Generic_IEnumerator_1_get_Current()), None, empty(), 1)

    else: 
        raise Exception("empty workflow metadata sheet")



def ArcWorkflow_toRows(workflow: ArcWorkflow) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def _arrow3644(__unit: None=None, workflow: Any=workflow) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3643(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3642(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3641(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    return to_rows_1("Workflow Person", of_seq(workflow.Contacts))

                return append(singleton(SparseRowModule_fromValues(to_enumerable(["WORKFLOW CONTACTS"]))), delay(_arrow3641))

            return append(to_rows(workflow), delay(_arrow3642))

        return append(singleton(SparseRowModule_fromValues(to_enumerable(["WORKFLOW"]))), delay(_arrow3643))

    return delay(_arrow3644)


def ArcWorkflow_toMetadataSheet(workflow: ArcWorkflow) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet("isa_workflow")
    def action(row_i: int, r: IEnumerable_1[tuple[int, str]], workflow: Any=workflow) -> None:
        SparseRowModule_writeToSheet(row_i + 1, r, sheet)

    iterate_indexed(action, ArcWorkflow_toRows(workflow))
    return sheet


def ArcWorkflow_fromMetadataSheet(sheet: FsWorksheet) -> ArcWorkflow:
    try: 
        return ArcWorkflow_fromRows(map(SparseRowModule_fromFsRow, sheet.Rows))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcWorkflow_toMetadataCollection(workflow: ArcWorkflow) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], workflow: Any=workflow) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    return map(mapping, ArcWorkflow_toRows(workflow))


def ArcWorkflow_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> ArcWorkflow:
    try: 
        return ArcWorkflow_fromRows(map(SparseRowModule_fromAllValues, collection))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcWorkflow_isMetadataSheetName(name: str) -> bool:
    return name == "isa_workflow"


def ArcWorkflow_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return ArcWorkflow_isMetadataSheetName(sheet.Name)


def ArcWorkflow_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return ArcWorkflow_isMetadataSheet(sheet)

    return try_find(predicate, doc.GetWorksheets())


def ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D(doc: FsWorkbook) -> ArcWorkflow:
    try: 
        workflow_metadata: ArcWorkflow
        match_value: FsWorksheet | None = ArcWorkflow_tryGetMetadataSheet(doc)
        if match_value is None:
            to_console(printf("Cannot retrieve metadata: Workflow file does not contain \"%s\" sheet."))("isa_workflow")
            workflow_metadata = ArcWorkflow.create(create_missing_identifier())

        else: 
            workflow_metadata = ArcWorkflow_fromMetadataSheet(match_value)

        datamap_sheet: DataMap | None = try_pick(try_from_fs_worksheet, doc.GetWorksheets())
        workflow_metadata.DataMap = datamap_sheet
        return workflow_metadata

    except Exception as err:
        arg_1: str = str(err)
        return to_fail(printf("Could not parse assay: \n%s"))(arg_1)



def ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(workflow: ArcWorkflow) -> FsWorkbook:
    doc: FsWorkbook = FsWorkbook()
    metadata_sheet: FsWorksheet = ArcWorkflow_toMetadataSheet(workflow)
    doc.AddWorksheet(metadata_sheet)
    return doc


def ARCtrl_ArcWorkflow__ArcWorkflow_ToFsWorkbook(this: ArcWorkflow) -> FsWorkbook:
    return ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(this)


__all__ = ["ArcWorkflow_fromRows", "ArcWorkflow_toRows", "ArcWorkflow_toMetadataSheet", "ArcWorkflow_fromMetadataSheet", "ArcWorkflow_toMetadataCollection", "ArcWorkflow_fromMetadataCollection", "ArcWorkflow_isMetadataSheetName", "ArcWorkflow_isMetadataSheet", "ArcWorkflow_tryGetMetadataSheet", "ARCtrl_ArcWorkflow__ArcWorkflow_fromFsWorkbook_Static_32154C9D", "ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E", "ARCtrl_ArcWorkflow__ArcWorkflow_ToFsWorkbook"]




from abc import abstractmethod
from typing import Protocol

class VersionInfo(Protocol):
    @property
    @abstractmethod
    def major(self) -> int:
        ...

    @property
    @abstractmethod
    def micro(self) -> int:
        ...

    @property
    @abstractmethod
    def minor(self) -> int:
        ...

    @property
    @abstractmethod
    def releaselevel(self) -> str:
        ...

    @property
    @abstractmethod
    def serial(self) -> int:
        ...





from __future__ import annotations
from typing import Any
from ...fable_library.list import FSharpList
from ...fable_library.range import range_big_int
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (to_list, map)
from ...fable_library.util import IEnumerable_1
from ..Cells.fs_cell import FsCell
from ..Cells.fs_cells_collection import FsCellsCollection
from ..fs_address import FsAddress
from .fs_range_address import FsRangeAddress
from .fs_range_base import (FsRangeBase, FsRangeBase_reflection)

def _expr216() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRangeColumn", None, FsRangeColumn, FsRangeBase_reflection())


class FsRangeColumn(FsRangeBase):
    def __init__(self, range_address: FsRangeAddress) -> None:
        super().__init__(range_address)
        pass

    @staticmethod
    def from_index(index: int) -> FsRangeColumn:
        return FsRangeColumn(FsRangeAddress(FsAddress(0, index), FsAddress(0, index)))

    @property
    def Index(self, __unit: None=None) -> int:
        self_1: FsRangeColumn = self
        return self_1.RangeAddress.FirstAddress.ColumnNumber

    @Index.setter
    def Index(self, i: int) -> None:
        self_1: FsRangeColumn = self
        self_1.RangeAddress.FirstAddress.ColumnNumber = i or 0
        self_1.RangeAddress.LastAddress.ColumnNumber = i or 0

    def Cell(self, row_index: int, cells_collection: FsCellsCollection) -> FsCell:
        return super().Cell(FsAddress((row_index - super().RangeAddress.FirstAddress.RowNumber) + 1, 1), cells_collection)

    def FirstCell(self, cells: FsCellsCollection) -> FsCell:
        return super().Cell(FsAddress(1, 1), cells)

    def Cells(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return super().Cells(cells_collection)

    @staticmethod
    def from_range_address(range_address: FsRangeAddress) -> FsRangeColumn:
        return FsRangeColumn(range_address)

    def Copy(self, __unit: None=None) -> FsRangeColumn:
        self_1: FsRangeColumn = self
        return FsRangeColumn(self_1.RangeAddress.Copy())

    @staticmethod
    def copy(range_column: FsRangeColumn) -> FsRangeColumn:
        return range_column.Copy()


FsRangeColumn_reflection = _expr216

def FsRangeColumn__ctor_6A2513BC(range_address: FsRangeAddress) -> FsRangeColumn:
    return FsRangeColumn(range_address)


def FsSpreadsheet_FsRangeAddress__FsRangeAddress_toRangeColumns_Static_6A2513BC(range_address: FsRangeAddress) -> IEnumerable_1[FsRangeColumn]:
    columns: FSharpList[int] = to_list(range_big_int(range_address.FirstAddress.ColumnNumber, 1, range_address.LastAddress.ColumnNumber))
    fst_row: int = range_address.FirstAddress.RowNumber or 0
    lst_row: int = range_address.LastAddress.RowNumber or 0
    def mapping(c: int, range_address: Any=range_address) -> FsRangeColumn:
        return FsRangeColumn(FsRangeAddress(FsAddress(fst_row, c), FsAddress(lst_row, c)))

    return map(mapping, columns)


__all__ = ["FsRangeColumn_reflection", "FsSpreadsheet_FsRangeAddress__FsRangeAddress_toRangeColumns_Static_6A2513BC"]




from __future__ import annotations
from typing import Any
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import LDNode

def _expr1474() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDOrganization", None, LDOrganization)


class LDOrganization:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/Organization"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def try_get_name_as_string(o: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = o.TryGetPropertyAsSingleton(LDOrganization.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(o: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = o.TryGetPropertyAsSingleton(LDOrganization.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + o.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + o.Id) + "`")


    @staticmethod
    def set_name_as_string(o: LDNode, n: str, context: LDContext | None=None) -> Any:
        return o.SetProperty(LDOrganization.name(), n, context)

    @staticmethod
    def gen_id(name: str) -> str:
        return clean(("#Organization_" + name) + "")

    @staticmethod
    def validate(o: LDNode, context: LDContext | None=None) -> bool:
        return o.HasProperty(LDOrganization.name(), context) if o.HasType(LDOrganization.schema_type(), context) else False

    @staticmethod
    def create(name: str, id: str | None=None, context: LDContext | None=None) -> LDNode:
        o: LDNode = LDNode(LDOrganization.gen_id(name) if (id is None) else id, [LDOrganization.schema_type()], None, context)
        o.SetProperty(LDOrganization.name(), name, context)
        return o


LDOrganization_reflection = _expr1474

__all__ = ["LDOrganization_reflection"]

